# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "network public-ip prefix create",
)
class Create(AAZCommand):
    """Create a public IP prefix resource.

    :example: Create a public IP prefix resource. (autogenerated)
        az network public-ip prefix create --length 28 --location westus2 --name MyPublicIPPrefix --resource-group MyResourceGroup
    """

    _aaz_info = {
        "version": "2022-09-01",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.network/publicipprefixes/{}", "2022-09-01"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.name = AAZStrArg(
            options=["-n", "--name"],
            help="The name of the public IP prefix.",
            required=True,
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.edge_zone = AAZStrArg(
            options=["--edge-zone"],
            help="The name of edge zone.",
        )
        _args_schema.location = AAZResourceLocationArg(
            help="Location. Values from: `az account list-locations`. You can configure the default location using `az configure --defaults location=<location>`.",
            fmt=AAZResourceLocationArgFormat(
                resource_group_arg="resource_group",
            ),
        )
        _args_schema.custom_ip_prefix_name = AAZStrArg(
            options=["--custom-ip-prefix-name"],
            help="A custom prefix from which the public prefix derived. If you'd like to cross subscription, please use Resource ID instead.",
        )
        _args_schema.length = AAZIntArg(
            options=["--length"],
            help="Length of the prefix (i.e. `XX.XX.XX.XX/<Length>`).",
        )
        _args_schema.version = AAZStrArg(
            options=["--version"],
            help="IP address type.  Allowed values: IPv4, IPv6.",
            default="IPv4",
            enum={"IPv4": "IPv4", "IPv6": "IPv6"},
        )
        _args_schema.tier = AAZStrArg(
            options=["--tier"],
            help="Tier of a public IP prefix SKU.",
            enum={"Global": "Global", "Regional": "Regional"},
        )
        _args_schema.tags = AAZDictArg(
            options=["--tags"],
            help="Space-separated tags: key[=value] [key[=value] ...]. Use \"\" to clear existing tags.",
        )
        _args_schema.zone = AAZListArg(
            options=["-z", "--zone"],
            help="Space-separated list of availability zones into which to provision the resource.  Allowed values: 1, 2, 3.",
        )

        tags = cls._args_schema.tags
        tags.Element = AAZStrArg()

        zone = cls._args_schema.zone
        zone.Element = AAZStrArg()

        # define Arg Group "ExtendedLocation"

        _args_schema = cls._args_schema
        _args_schema.type = AAZStrArg(
            options=["--type"],
            arg_group="ExtendedLocation",
            help="The type of the extended location.",
            enum={"EdgeZone": "EdgeZone"},
        )

        # define Arg Group "Parameters"

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.ip_tags_list = AAZListArg(
            options=["--ip-tags-list"],
            arg_group="Properties",
            help="The list of tags associated with the public IP prefix.",
        )

        ip_tags_list = cls._args_schema.ip_tags_list
        ip_tags_list.Element = AAZObjectArg()

        _element = cls._args_schema.ip_tags_list.Element
        _element.ip_tag_type = AAZStrArg(
            options=["ip-tag-type"],
            help="The IP tag type. Example: FirstPartyUsage.",
        )
        _element.tag = AAZStrArg(
            options=["tag"],
            help="The value of the IP tag associated with the public IP. Example: SQL.",
        )

        # define Arg Group "Sku"

        _args_schema = cls._args_schema
        _args_schema.sku = AAZStrArg(
            options=["--sku"],
            arg_group="Sku",
            help="Name of a public IP prefix SKU.",
            default="Standard",
            enum={"Standard": "Standard"},
        )
        return cls._args_schema

    _args_sub_resource_create = None

    @classmethod
    def _build_args_sub_resource_create(cls, _schema):
        if cls._args_sub_resource_create is not None:
            _schema.id = cls._args_sub_resource_create.id
            return

        cls._args_sub_resource_create = AAZObjectArg()

        sub_resource_create = cls._args_sub_resource_create
        sub_resource_create.id = AAZStrArg(
            options=["id"],
            help="Resource ID.",
        )

        _schema.id = cls._args_sub_resource_create.id

    def _execute_operations(self):
        self.pre_operations()
        yield self.PublicIPPrefixesCreateOrUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class PublicIPPrefixesCreateOrUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "location"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200, 201]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200_201,
                    self.on_error,
                    lro_options={"final-state-via": "location"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Network/publicIPPrefixes/{publicIpPrefixName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PUT"

        @property
        def error_format(self):
            return "ODataV4Format"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "publicIpPrefixName", self.ctx.args.name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2022-09-01",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("extendedLocation", AAZObjectType)
            _builder.set_prop("location", AAZStrType, ".location")
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})
            _builder.set_prop("sku", AAZObjectType)
            _builder.set_prop("tags", AAZDictType, ".tags")
            _builder.set_prop("zones", AAZListType, ".zone")

            extended_location = _builder.get(".extendedLocation")
            if extended_location is not None:
                extended_location.set_prop("name", AAZStrType, ".edge_zone")
                extended_location.set_prop("type", AAZStrType, ".type")

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("customIPPrefix", AAZObjectType)
                properties.set_prop("ipTags", AAZListType, ".ip_tags_list")
                properties.set_prop("prefixLength", AAZIntType, ".length")
                properties.set_prop("publicIPAddressVersion", AAZStrType, ".version")

            custom_ip_prefix = _builder.get(".properties.customIPPrefix")
            if custom_ip_prefix is not None:
                custom_ip_prefix.set_prop("id", AAZStrType, ".custom_ip_prefix_name")

            ip_tags = _builder.get(".properties.ipTags")
            if ip_tags is not None:
                ip_tags.set_elements(AAZObjectType, ".")

            _elements = _builder.get(".properties.ipTags[]")
            if _elements is not None:
                _elements.set_prop("ipTagType", AAZStrType, ".ip_tag_type")
                _elements.set_prop("tag", AAZStrType, ".tag")

            sku = _builder.get(".sku")
            if sku is not None:
                sku.set_prop("name", AAZStrType, ".sku")
                sku.set_prop("tier", AAZStrType, ".tier")

            tags = _builder.get(".tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            zones = _builder.get(".zones")
            if zones is not None:
                zones.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)

        def on_200_201(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200_201
            )

        _schema_on_200_201 = None

        @classmethod
        def _build_schema_on_200_201(cls):
            if cls._schema_on_200_201 is not None:
                return cls._schema_on_200_201

            cls._schema_on_200_201 = AAZObjectType()

            _schema_on_200_201 = cls._schema_on_200_201
            _schema_on_200_201.etag = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.extended_location = AAZObjectType(
                serialized_name="extendedLocation",
            )
            _schema_on_200_201.id = AAZStrType()
            _schema_on_200_201.location = AAZStrType()
            _schema_on_200_201.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200_201.sku = AAZObjectType()
            _schema_on_200_201.tags = AAZDictType()
            _schema_on_200_201.type = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200_201.zones = AAZListType()

            extended_location = cls._schema_on_200_201.extended_location
            extended_location.name = AAZStrType()
            extended_location.type = AAZStrType()

            properties = cls._schema_on_200_201.properties
            properties.custom_ip_prefix = AAZObjectType(
                serialized_name="customIPPrefix",
            )
            _CreateHelper._build_schema_sub_resource_read(properties.custom_ip_prefix)
            properties.ip_prefix = AAZStrType(
                serialized_name="ipPrefix",
                flags={"read_only": True},
            )
            properties.ip_tags = AAZListType(
                serialized_name="ipTags",
            )
            properties.load_balancer_frontend_ip_configuration = AAZObjectType(
                serialized_name="loadBalancerFrontendIpConfiguration",
            )
            _CreateHelper._build_schema_sub_resource_read(properties.load_balancer_frontend_ip_configuration)
            properties.nat_gateway = AAZObjectType(
                serialized_name="natGateway",
            )
            properties.prefix_length = AAZIntType(
                serialized_name="prefixLength",
            )
            properties.provisioning_state = AAZStrType(
                serialized_name="provisioningState",
                flags={"read_only": True},
            )
            properties.public_ip_address_version = AAZStrType(
                serialized_name="publicIPAddressVersion",
            )
            properties.public_ip_addresses = AAZListType(
                serialized_name="publicIPAddresses",
                flags={"read_only": True},
            )
            properties.resource_guid = AAZStrType(
                serialized_name="resourceGuid",
                flags={"read_only": True},
            )

            ip_tags = cls._schema_on_200_201.properties.ip_tags
            ip_tags.Element = AAZObjectType()

            _element = cls._schema_on_200_201.properties.ip_tags.Element
            _element.ip_tag_type = AAZStrType(
                serialized_name="ipTagType",
            )
            _element.tag = AAZStrType()

            nat_gateway = cls._schema_on_200_201.properties.nat_gateway
            nat_gateway.etag = AAZStrType(
                flags={"read_only": True},
            )
            nat_gateway.id = AAZStrType()
            nat_gateway.location = AAZStrType()
            nat_gateway.name = AAZStrType(
                flags={"read_only": True},
            )
            nat_gateway.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            nat_gateway.sku = AAZObjectType()
            nat_gateway.tags = AAZDictType()
            nat_gateway.type = AAZStrType(
                flags={"read_only": True},
            )
            nat_gateway.zones = AAZListType()

            properties = cls._schema_on_200_201.properties.nat_gateway.properties
            properties.idle_timeout_in_minutes = AAZIntType(
                serialized_name="idleTimeoutInMinutes",
            )
            properties.provisioning_state = AAZStrType(
                serialized_name="provisioningState",
                flags={"read_only": True},
            )
            properties.public_ip_addresses = AAZListType(
                serialized_name="publicIpAddresses",
            )
            properties.public_ip_prefixes = AAZListType(
                serialized_name="publicIpPrefixes",
            )
            properties.resource_guid = AAZStrType(
                serialized_name="resourceGuid",
                flags={"read_only": True},
            )
            properties.subnets = AAZListType(
                flags={"read_only": True},
            )

            public_ip_addresses = cls._schema_on_200_201.properties.nat_gateway.properties.public_ip_addresses
            public_ip_addresses.Element = AAZObjectType()
            _CreateHelper._build_schema_sub_resource_read(public_ip_addresses.Element)

            public_ip_prefixes = cls._schema_on_200_201.properties.nat_gateway.properties.public_ip_prefixes
            public_ip_prefixes.Element = AAZObjectType()
            _CreateHelper._build_schema_sub_resource_read(public_ip_prefixes.Element)

            subnets = cls._schema_on_200_201.properties.nat_gateway.properties.subnets
            subnets.Element = AAZObjectType()
            _CreateHelper._build_schema_sub_resource_read(subnets.Element)

            sku = cls._schema_on_200_201.properties.nat_gateway.sku
            sku.name = AAZStrType()

            tags = cls._schema_on_200_201.properties.nat_gateway.tags
            tags.Element = AAZStrType()

            zones = cls._schema_on_200_201.properties.nat_gateway.zones
            zones.Element = AAZStrType()

            public_ip_addresses = cls._schema_on_200_201.properties.public_ip_addresses
            public_ip_addresses.Element = AAZObjectType()

            _element = cls._schema_on_200_201.properties.public_ip_addresses.Element
            _element.id = AAZStrType()

            sku = cls._schema_on_200_201.sku
            sku.name = AAZStrType()
            sku.tier = AAZStrType()

            tags = cls._schema_on_200_201.tags
            tags.Element = AAZStrType()

            zones = cls._schema_on_200_201.zones
            zones.Element = AAZStrType()

            return cls._schema_on_200_201


class _CreateHelper:
    """Helper class for Create"""

    @classmethod
    def _build_schema_sub_resource_create(cls, _builder):
        if _builder is None:
            return
        _builder.set_prop("id", AAZStrType, ".id")

    _schema_sub_resource_read = None

    @classmethod
    def _build_schema_sub_resource_read(cls, _schema):
        if cls._schema_sub_resource_read is not None:
            _schema.id = cls._schema_sub_resource_read.id
            return

        cls._schema_sub_resource_read = _schema_sub_resource_read = AAZObjectType()

        sub_resource_read = _schema_sub_resource_read
        sub_resource_read.id = AAZStrType()

        _schema.id = cls._schema_sub_resource_read.id


__all__ = ["Create"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.azclierror import UnknownError
from azure.cli.core.commands import LongRunningOperation
from azure.cli.command_modules.acs.azurecontainerstorage._consts import (
    CONST_ACSTOR_K8S_EXTENSION_NAME,
    CONST_EXT_INSTALLATION_NAME,
    CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME,
    CONST_K8S_EXTENSION_CUSTOM_MOD_NAME,
    CONST_STORAGE_POOL_DEFAULT_SIZE,
    CONST_STORAGE_POOL_DEFAULT_SIZE_ESAN,
    CONST_STORAGE_POOL_OPTION_NVME,
    CONST_STORAGE_POOL_OPTION_SSD,
    CONST_STORAGE_POOL_SKU_PREMIUM_LRS,
    CONST_ACSTOR_ALL,
    CONST_STORAGE_POOL_TYPE_AZURE_DISK,
    CONST_STORAGE_POOL_TYPE_ELASTIC_SAN,
    CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK,
)
from azure.cli.command_modules.acs.azurecontainerstorage._helpers import (
    get_k8s_extension_module,
    get_current_resource_value_args,
    get_desired_resource_value_args,
    get_initial_resource_value_args,
    perform_role_operations_on_managed_rg,
    validate_storagepool_creation,
)
from knack.log import get_logger

logger = get_logger(__name__)


def perform_enable_azure_container_storage(  # pylint: disable=too-many-statements,too-many-locals,too-many-branches
    cmd,
    subscription_id,
    resource_group,
    cluster_name,
    node_resource_group,
    kubelet_identity_object_id,
    storage_pool_name,
    storage_pool_type,
    storage_pool_size,
    storage_pool_sku,
    storage_pool_option,
    acstor_nodepool_skus,
    is_cluster_create,
    is_extension_installed=False,
    is_azureDisk_enabled=False,
    is_elasticSan_enabled=False,
    is_ephemeralDisk_localssd_enabled=False,
    is_ephemeralDisk_nvme_enabled=False,
    current_core_value=None,
):
    # Step 1: Validate if storagepool could be created.
    # Depends on the following:
    #   1a: Grant AKS cluster's node identity the following
    #       roles on the AKS managed resource group:
    #       Azure Container Storage Operator.
    #       Ensure grant was successful if creation of
    #       Elastic SAN storagepool is requested.
    validate_storagepool_creation(
        cmd,
        subscription_id,
        node_resource_group,
        kubelet_identity_object_id,
        storage_pool_type,
    )

    # Step 3: Configure the storagepool parameters
    config_settings = []
    if storage_pool_name is None:
        storage_pool_name = storage_pool_type.lower()
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
            storage_pool_name = storage_pool_type.lower() + "-" + storage_pool_option.lower()
    if storage_pool_size is None:
        storage_pool_size = CONST_STORAGE_POOL_DEFAULT_SIZE_ESAN if \
            storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN else \
            CONST_STORAGE_POOL_DEFAULT_SIZE

    azure_disk_enabled = is_azureDisk_enabled if is_extension_installed else False
    elastic_san_enabled = is_elasticSan_enabled if is_extension_installed else False
    ephemeral_disk_nvme_enabled = is_ephemeralDisk_nvme_enabled if is_extension_installed else False
    ephemeral_disk_localssd_enabled = is_ephemeralDisk_localssd_enabled if is_extension_installed else False

    epheremaldisk_type = ""
    if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
        if storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME:
            ephemeral_disk_nvme_enabled = True
        elif storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD:
            ephemeral_disk_localssd_enabled = True
        epheremaldisk_type = storage_pool_option.lower()
    else:
        if storage_pool_sku is None:
            storage_pool_sku = CONST_STORAGE_POOL_SKU_PREMIUM_LRS
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN:
            config_settings.append({"global.cli.storagePool.elasticSan.sku": storage_pool_sku})
            elastic_san_enabled = True
        elif storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK:
            config_settings.append({"global.cli.storagePool.azureDisk.sku": storage_pool_sku})
            azure_disk_enabled = True

    config_settings.extend(
        [
            {"global.cli.activeControl": True},
            {"global.cli.storagePool.install.create": True},
            {"global.cli.storagePool.install.name": storage_pool_name},
            {"global.cli.storagePool.install.size": storage_pool_size},
            {"global.cli.storagePool.install.type": storage_pool_type},
            {"global.cli.storagePool.install.diskType": epheremaldisk_type},
            {"global.cli.storagePool.azureDisk.enabled": azure_disk_enabled},
            {"global.cli.storagePool.elasticSan.enabled": elastic_san_enabled},
            {"global.cli.storagePool.ephemeralDisk.nvme.enabled": ephemeral_disk_nvme_enabled},
            {"global.cli.storagePool.ephemeralDisk.temp.enabled": ephemeral_disk_localssd_enabled},
            # Always set cli.storagePool.disable.type to empty
            # and cli.storagePool.disable.validation to False
            # during enable operation so that any older disable
            # operation doesn't interfere with the current state.
            {"global.cli.storagePool.disable.validation": False},
            {"global.cli.storagePool.disable.type": ""},
            {"global.cli.storagePool.disable.active": False},
            # Resetting older cluster values from previous ARC version.
            {"cli.storagePool.type": ""},
            {"cli.storagePool.ephemeralDisk.diskType": ""},
        ]
    )

    if is_extension_installed:
        resource_args = get_desired_resource_value_args(
            storage_pool_type,
            storage_pool_option,
            current_core_value,
            is_azureDisk_enabled,
            is_elasticSan_enabled,
            is_ephemeralDisk_localssd_enabled,
            is_ephemeralDisk_nvme_enabled,
            acstor_nodepool_skus,
            True,
        )

    else:
        resource_args = get_initial_resource_value_args(
            storage_pool_type,
            storage_pool_option,
            acstor_nodepool_skus,
        )

    config_settings.extend(resource_args)

    # Step 5: Install the k8s_extension 'microsoft.azurecontainerstorage'
    client_factory = get_k8s_extension_module(CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME)
    client = client_factory.cf_k8s_extension_operation(cmd.cli_ctx)

    k8s_extension_custom_mod = get_k8s_extension_module(CONST_K8S_EXTENSION_CUSTOM_MOD_NAME)
    update_settings = [
        {"global.cli.activeControl": True},
        {"global.cli.storagePool.install.create": False},
        {"global.cli.storagePool.install.name": ""},
        {"global.cli.storagePool.install.size": ""},
        {"global.cli.storagePool.install.type": ""},
        {"global.cli.storagePool.install.diskType": ""},
    ]
    try:
        if is_extension_installed:
            result = k8s_extension_custom_mod.update_k8s_extension(
                cmd,
                client,
                resource_group,
                cluster_name,
                CONST_EXT_INSTALLATION_NAME,
                "managedClusters",
                configuration_settings=config_settings,
                yes=True,
                no_wait=False,
            )
            op_text = "Azure Container Storage successfully updated"
        else:
            result = k8s_extension_custom_mod.create_k8s_extension(
                cmd,
                client,
                resource_group,
                cluster_name,
                CONST_EXT_INSTALLATION_NAME,
                "managedClusters",
                CONST_ACSTOR_K8S_EXTENSION_NAME,
                auto_upgrade_minor_version=True,
                release_train="stable",
                scope="cluster",
                release_namespace="acstor",
                configuration_settings=config_settings,
            )
            op_text = "Azure Container Storage successfully installed"

        long_op_result = LongRunningOperation(cmd.cli_ctx)(result)
        if long_op_result.provisioning_state == "Succeeded":
            logger.warning(op_text)
    except Exception as ex:  # pylint: disable=broad-except
        if is_cluster_create:
            logger.error("Azure Container Storage failed to install.\nError: %s", ex)
            logger.warning(
                "AKS cluster is created. "
                "Please run `az aks update` along with `--enable-azure-container-storage` "
                "to enable Azure Container Storage."
            )
        else:
            if is_extension_installed:
                logger.error(
                    "AKS update to enable Azure Container Storage pool type %s failed. \n"
                    " Error: %s. Resetting cluster state.", storage_pool_type, ex
                )

                # If enabling of storagepool type in Azure Container Storage failed,
                # define the existing resource values for ioEngine and hugepages for
                # resetting the cluster state.
                resource_args = get_current_resource_value_args(
                    is_azureDisk_enabled,
                    is_elasticSan_enabled,
                    is_ephemeralDisk_localssd_enabled,
                    is_ephemeralDisk_nvme_enabled,
                    current_core_value,
                )

                update_settings.extend(resource_args)
                # Also, unset the type of storagepool which was supposed to enabled.
                if storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK:
                    update_settings.append({"global.cli.storagePool.azureDisk.enabled": False})
                elif storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN:
                    update_settings.append({"global.cli.storagePool.elasticSan.enabled": False})
                elif storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
                    if storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME:
                        update_settings.append({"global.cli.storagePool.ephemeralDisk.nvme.enabled": False})
                    elif storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD:
                        update_settings.append({"global.cli.storagePool.ephemeralDisk.temp.enabled": False})
            else:
                logger.error("AKS update to enable Azure Container Storage failed.\nError: %s", ex)

    k8s_extension_custom_mod.update_k8s_extension(
        cmd,
        client,
        resource_group,
        cluster_name,
        CONST_EXT_INSTALLATION_NAME,
        "managedClusters",
        configuration_settings=update_settings,
        yes=True,
        no_wait=True,
    )


def perform_disable_azure_container_storage(  # pylint: disable=too-many-statements,too-many-locals,too-many-branches
    cmd,
    subscription_id,
    resource_group,
    cluster_name,
    node_resource_group,
    kubelet_identity_object_id,
    perform_validation,
    storage_pool_type,
    storage_pool_option,
    is_elasticSan_enabled,
    is_azureDisk_enabled,
    is_ephemeralDisk_localssd_enabled,
    is_ephemeralDisk_nvme_enabled,
    current_core_value,
):
    client_factory = get_k8s_extension_module(CONST_K8S_EXTENSION_CLIENT_FACTORY_MOD_NAME)
    client = client_factory.cf_k8s_extension_operation(cmd.cli_ctx)
    k8s_extension_custom_mod = get_k8s_extension_module(CONST_K8S_EXTENSION_CUSTOM_MOD_NAME)
    no_wait_delete_op = False

    azure_disk_enabled = is_azureDisk_enabled
    elastic_san_enabled = is_elasticSan_enabled
    ephemeral_disk_nvme_enabled = is_ephemeralDisk_nvme_enabled
    ephemeral_disk_localssd_enabled = is_ephemeralDisk_localssd_enabled

    # Ensure that all the install storagepool fields are turned off.
    reset_install_settings = [
        {"global.cli.activeControl": True},
        {"global.cli.storagePool.install.create": False},
        {"global.cli.storagePool.install.name": ""},
        {"global.cli.storagePool.install.size": ""},
        {"global.cli.storagePool.install.type": ""},
        {"global.cli.storagePool.install.diskType": ""},
    ]

    pool_option = ""
    if storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
        if storage_pool_option is not None:
            pool_option = storage_pool_option.lower()
        else:
            if is_ephemeralDisk_nvme_enabled:
                pool_option = CONST_STORAGE_POOL_OPTION_NVME.lower()
            elif is_ephemeralDisk_localssd_enabled:
                pool_option = CONST_STORAGE_POOL_OPTION_SSD.lower()

    # Step 1: Perform validation if accepted by user
    if perform_validation:
        config_settings = [
            {"global.cli.storagePool.disable.validation": True},
            {"global.cli.storagePool.disable.type": storage_pool_type},
            # Set these values to ensure cluster state incase of
            # a cluster where cli operation has not yet run or older
            # version of charts were installed.
            {"global.cli.storagePool.azureDisk.enabled": azure_disk_enabled},
            {"global.cli.storagePool.elasticSan.enabled": elastic_san_enabled},
            {"global.cli.storagePool.ephemeralDisk.nvme.enabled": ephemeral_disk_nvme_enabled},
            {"global.cli.storagePool.ephemeralDisk.temp.enabled": ephemeral_disk_localssd_enabled},
        ]

        config_settings.extend(reset_install_settings)
        config_settings.append({"global.cli.storagePool.disable.diskType": pool_option.lower()})

        try:
            update_result = k8s_extension_custom_mod.update_k8s_extension(
                cmd,
                client,
                resource_group,
                cluster_name,
                CONST_EXT_INSTALLATION_NAME,
                "managedClusters",
                configuration_settings=config_settings,
                yes=True,
                no_wait=False,
            )

            update_long_op_result = LongRunningOperation(cmd.cli_ctx)(update_result)
            if update_long_op_result.provisioning_state == "Succeeded":
                logger.warning("Validation succeeded. Continuing Azure Container Storage disable operation...")

            # Since, pre uninstall validation will ensure deletion of storagepools,
            # we don't need to long wait while performing the delete operation.
            # Setting no_wait_delete_op = True.
            # Only relevant when we are uninstalling Azure Container Storage completely.
            if storage_pool_type == CONST_ACSTOR_ALL:
                no_wait_delete_op = True
        except Exception as ex:  # pylint: disable=broad-except
            config_settings = [
                {"global.cli.storagePool.disable.validation": False},
                {"global.cli.storagePool.disable.type": ""},
                {"global.cli.storagePool.disable.diskType": ""},
                # Set these values to ensure cluster state incase of
                # a cluster where cli operation has not yet run or older
                # version of charts were installed.
                {"global.cli.storagePool.azureDisk.enabled": azure_disk_enabled},
                {"global.cli.storagePool.elasticSan.enabled": elastic_san_enabled},
                {"global.cli.storagePool.ephemeralDisk.nvme.enabled": ephemeral_disk_nvme_enabled},
                {"global.cli.storagePool.ephemeralDisk.temp.enabled": ephemeral_disk_localssd_enabled},
            ]

            config_settings.extend(reset_install_settings)

            err_msg = (
                "Validation failed. "
                "Please ensure that storagepools are not being used. "
                "Unable to perform disable Azure Container Storage operation. "
                "Resetting cluster state."
            )

            if storage_pool_type != CONST_ACSTOR_ALL:
                err_msg = (
                    "Validation failed. "
                    f"Please ensure that storagepools of type {storage_pool_type} are not being used. "
                    f"Unable to perform disable Azure Container Storage operation. "
                    "Resetting cluster state."
                )

            k8s_extension_custom_mod.update_k8s_extension(
                cmd,
                client,
                resource_group,
                cluster_name,
                CONST_EXT_INSTALLATION_NAME,
                "managedClusters",
                configuration_settings=config_settings,
                yes=True,
                no_wait=True,
            )

            if "pre-upgrade hooks failed" in str(ex):
                raise UnknownError(err_msg) from ex
            raise UnknownError(
                "Validation failed. Unable to perform Azure Container Storage operation. Resetting cluster state."
            ) from ex

    # Step 2: If the extension is installed and validation succeeded or skipped,
    # if we want to uninstall Azure Container Storage completely, then call
    # delete_k8s_extension. Else, if we want to disable a storagepool type,
    # we will perform another update_k8s_extension operation on the cluster.

    if storage_pool_type == CONST_ACSTOR_ALL:
        # Uninstallation operation
        try:
            delete_op_result = k8s_extension_custom_mod.delete_k8s_extension(
                cmd,
                client,
                resource_group,
                cluster_name,
                CONST_EXT_INSTALLATION_NAME,
                "managedClusters",
                yes=True,
                no_wait=no_wait_delete_op,
            )

            if not no_wait_delete_op:
                LongRunningOperation(cmd.cli_ctx)(delete_op_result)
        except Exception as delete_ex:
            raise UnknownError(
                "Failure observed while disabling Azure Container Storage."
            ) from delete_ex

        logger.warning("Azure Container Storage has been disabled.")

        # Revoke role assignments irrespective of whether ElasticSAN
        # type was enabled on the cluster to handle older clusters where
        # the role assignments were done for all storagepool type during
        # installation of Azure Container Storage.
        perform_role_operations_on_managed_rg(
            cmd,
            subscription_id,
            node_resource_group,
            kubelet_identity_object_id,
            False
        )
    else:
        # Disabling a particular type of storagepool.
        if storage_pool_type == CONST_STORAGE_POOL_TYPE_AZURE_DISK:
            azure_disk_enabled = False
        elif storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN:
            elastic_san_enabled = False
        elif storage_pool_type == CONST_STORAGE_POOL_TYPE_EPHEMERAL_DISK:
            if storage_pool_option == CONST_STORAGE_POOL_OPTION_NVME:
                ephemeral_disk_nvme_enabled = False
            elif storage_pool_option == CONST_STORAGE_POOL_OPTION_SSD:
                ephemeral_disk_localssd_enabled = False
            elif storage_pool_option == CONST_ACSTOR_ALL:
                ephemeral_disk_nvme_enabled = False
                ephemeral_disk_localssd_enabled = False

        config_settings = [
            {"global.cli.storagePool.disable.validation": False},
            {"global.cli.storagePool.disable.type": storage_pool_type},
            {"global.cli.storagePool.disable.active": True},
            {"global.cli.storagePool.disable.diskType": pool_option.lower()},
            # Set these values to ensure cluster state incase of
            # a cluster where cli operation has not yet run or older
            # version of charts were installed.
            {"global.cli.storagePool.azureDisk.enabled": azure_disk_enabled},
            {"global.cli.storagePool.elasticSan.enabled": elastic_san_enabled},
            {"global.cli.storagePool.ephemeralDisk.nvme.enabled": ephemeral_disk_nvme_enabled},
            {"global.cli.storagePool.ephemeralDisk.temp.enabled": ephemeral_disk_localssd_enabled},
        ]

        config_settings.extend(reset_install_settings)

        # Define the new resource values for ioEngine and hugepages.
        resource_args = get_desired_resource_value_args(
            storage_pool_type,
            storage_pool_option,
            current_core_value,
            is_azureDisk_enabled,
            is_elasticSan_enabled,
            is_ephemeralDisk_localssd_enabled,
            is_ephemeralDisk_nvme_enabled,
            None,
            False,
        )
        config_settings.extend(resource_args)
        # Creating the set of config settings which
        # are required to demarcate that the disabling
        # process is completed. This config variable
        # will be used after the disabling operation is completed.
        update_settings = [
            {"global.cli.activeControl": True},
            {"global.cli.storagePool.disable.validation": False},
            {"global.cli.storagePool.disable.type": ""},
            {"global.cli.storagePool.disable.active": False},
            {"global.cli.storagePool.disable.diskType": ""},
        ]

        update_settings.extend(reset_install_settings)
        disable_op_failure = False
        try:
            disable_op_result = k8s_extension_custom_mod.update_k8s_extension(
                cmd,
                client,
                resource_group,
                cluster_name,
                CONST_EXT_INSTALLATION_NAME,
                "managedClusters",
                configuration_settings=config_settings,
                yes=True,
                no_wait=False,
            )
            LongRunningOperation(cmd.cli_ctx)(disable_op_result)

            # Revoke role assignments if we are disabling ElasticSAN storagepool type.
            if storage_pool_type == CONST_STORAGE_POOL_TYPE_ELASTIC_SAN and is_elasticSan_enabled:
                perform_role_operations_on_managed_rg(
                    cmd,
                    subscription_id,
                    node_resource_group,
                    kubelet_identity_object_id,
                    False
                )
        except Exception as disable_ex:  # pylint: disable=broad-except
            logger.error(
                "Failure observed while disabling Azure Container Storage storagepool type: %s.\nError: %s"
                "Resetting cluster state.", storage_pool_type, disable_ex
            )

            # If disabling type of storagepool in Azure Container Storage failed,
            # define the existing resource values for ioEngine and hugepages for
            # resetting the cluster state.
            resource_args = get_current_resource_value_args(
                is_azureDisk_enabled,
                is_elasticSan_enabled,
                is_ephemeralDisk_localssd_enabled,
                is_ephemeralDisk_nvme_enabled,
                current_core_value,
            )

            update_settings.extend(resource_args)
            # Revert back to storagepool type states which was supposed to disabled.
            azure_disk_enabled = is_azureDisk_enabled
            elastic_san_enabled = is_elasticSan_enabled
            ephemeral_disk_nvme_enabled = is_ephemeralDisk_nvme_enabled
            ephemeral_disk_localssd_enabled = is_ephemeralDisk_localssd_enabled
            disable_op_failure = True

        # Set the types of storagepool type states in the cluster
        # based on whether the previous operation succeeded or failed.
        update_settings.extend(
            [
                {"global.cli.storagePool.azureDisk.enabled": azure_disk_enabled},
                {"global.cli.storagePool.elasticSan.enabled": elastic_san_enabled},
                {"global.cli.storagePool.ephemeralDisk.nvme.enabled": ephemeral_disk_nvme_enabled},
                {"global.cli.storagePool.ephemeralDisk.temp.enabled": ephemeral_disk_localssd_enabled},
            ]
        )
        # Since we are just resetting the cluster state,
        # we are going to perform a non waiting operation.
        k8s_extension_custom_mod.update_k8s_extension(
            cmd,
            client,
            resource_group,
            cluster_name,
            CONST_EXT_INSTALLATION_NAME,
            "managedClusters",
            configuration_settings=update_settings,
            yes=True,
            no_wait=True,
        )

        if not disable_op_failure:
            logger.warning("Azure Container Storage storagepool type %s has been disabled.", storage_pool_type)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "security security-connector show",
    is_preview=True,
)
class Show(AAZCommand):
    """Get details of a specific security connector
    """

    _aaz_info = {
        "version": "2023-10-01-preview",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.security/securityconnectors/{}", "2023-10-01-preview"],
        ]
    }

    def _handler(self, command_args):
        super()._handler(command_args)
        self._execute_operations()
        return self._output()

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )
        _args_schema.security_connector_name = AAZStrArg(
            options=["-n", "--name", "--security-connector-name"],
            help="The security connector name.",
            required=True,
            id_part="name",
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        self.SecurityConnectorsGet(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class SecurityConnectorsGet(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [200]:
                return self.on_200(session)

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.Security/securityConnectors/{securityConnectorName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "GET"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "securityConnectorName", self.ctx.args.security_connector_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2023-10-01-preview",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()

            _schema_on_200 = cls._schema_on_200
            _schema_on_200.etag = AAZStrType()
            _schema_on_200.id = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.kind = AAZStrType()
            _schema_on_200.location = AAZStrType()
            _schema_on_200.name = AAZStrType(
                flags={"read_only": True},
            )
            _schema_on_200.properties = AAZObjectType(
                flags={"client_flatten": True},
            )
            _schema_on_200.system_data = AAZObjectType(
                serialized_name="systemData",
                flags={"read_only": True},
            )
            _schema_on_200.tags = AAZDictType()
            _schema_on_200.type = AAZStrType(
                flags={"read_only": True},
            )

            properties = cls._schema_on_200.properties
            properties.environment_data = AAZObjectType(
                serialized_name="environmentData",
            )
            properties.environment_name = AAZStrType(
                serialized_name="environmentName",
            )
            properties.hierarchy_identifier = AAZStrType(
                serialized_name="hierarchyIdentifier",
            )
            properties.hierarchy_identifier_trial_end_date = AAZStrType(
                serialized_name="hierarchyIdentifierTrialEndDate",
                flags={"read_only": True},
            )
            properties.offerings = AAZListType()

            environment_data = cls._schema_on_200.properties.environment_data
            environment_data.environment_type = AAZStrType(
                serialized_name="environmentType",
                flags={"required": True},
            )

            disc_aws_account = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "AwsAccount")
            disc_aws_account.account_name = AAZStrType(
                serialized_name="accountName",
                flags={"read_only": True},
            )
            disc_aws_account.organizational_data = AAZObjectType(
                serialized_name="organizationalData",
            )
            disc_aws_account.regions = AAZListType()
            disc_aws_account.scan_interval = AAZIntType(
                serialized_name="scanInterval",
            )

            organizational_data = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "AwsAccount").organizational_data
            organizational_data.organization_membership_type = AAZStrType(
                serialized_name="organizationMembershipType",
                flags={"required": True},
            )

            disc_member = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "AwsAccount").organizational_data.discriminate_by("organization_membership_type", "Member")
            disc_member.parent_hierarchy_id = AAZStrType(
                serialized_name="parentHierarchyId",
            )

            disc_organization = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "AwsAccount").organizational_data.discriminate_by("organization_membership_type", "Organization")
            disc_organization.excluded_account_ids = AAZListType(
                serialized_name="excludedAccountIds",
            )
            disc_organization.stackset_name = AAZStrType(
                serialized_name="stacksetName",
            )

            excluded_account_ids = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "AwsAccount").organizational_data.discriminate_by("organization_membership_type", "Organization").excluded_account_ids
            excluded_account_ids.Element = AAZStrType()

            regions = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "AwsAccount").regions
            regions.Element = AAZStrType()

            disc_gcp_project = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "GcpProject")
            disc_gcp_project.organizational_data = AAZObjectType(
                serialized_name="organizationalData",
            )
            disc_gcp_project.project_details = AAZObjectType(
                serialized_name="projectDetails",
            )
            disc_gcp_project.scan_interval = AAZIntType(
                serialized_name="scanInterval",
            )

            organizational_data = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "GcpProject").organizational_data
            organizational_data.organization_membership_type = AAZStrType(
                serialized_name="organizationMembershipType",
                flags={"required": True},
            )

            disc_member = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "GcpProject").organizational_data.discriminate_by("organization_membership_type", "Member")
            disc_member.management_project_number = AAZStrType(
                serialized_name="managementProjectNumber",
            )
            disc_member.parent_hierarchy_id = AAZStrType(
                serialized_name="parentHierarchyId",
            )

            disc_organization = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "GcpProject").organizational_data.discriminate_by("organization_membership_type", "Organization")
            disc_organization.excluded_project_numbers = AAZListType(
                serialized_name="excludedProjectNumbers",
            )
            disc_organization.organization_name = AAZStrType(
                serialized_name="organizationName",
                flags={"read_only": True},
            )
            disc_organization.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            disc_organization.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            excluded_project_numbers = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "GcpProject").organizational_data.discriminate_by("organization_membership_type", "Organization").excluded_project_numbers
            excluded_project_numbers.Element = AAZStrType()

            project_details = cls._schema_on_200.properties.environment_data.discriminate_by("environment_type", "GcpProject").project_details
            project_details.project_id = AAZStrType(
                serialized_name="projectId",
            )
            project_details.project_name = AAZStrType(
                serialized_name="projectName",
                flags={"read_only": True},
            )
            project_details.project_number = AAZStrType(
                serialized_name="projectNumber",
            )
            project_details.workload_identity_pool_id = AAZStrType(
                serialized_name="workloadIdentityPoolId",
                flags={"read_only": True},
            )

            offerings = cls._schema_on_200.properties.offerings
            offerings.Element = AAZObjectType()

            _element = cls._schema_on_200.properties.offerings.Element
            _element.description = AAZStrType(
                flags={"read_only": True},
            )
            _element.offering_type = AAZStrType(
                serialized_name="offeringType",
                flags={"required": True},
            )

            disc_cspm_monitor_aws = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "CspmMonitorAws")
            disc_cspm_monitor_aws.native_cloud_connection = AAZObjectType(
                serialized_name="nativeCloudConnection",
            )

            native_cloud_connection = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "CspmMonitorAws").native_cloud_connection
            native_cloud_connection.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            disc_cspm_monitor_gcp = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "CspmMonitorGcp")
            disc_cspm_monitor_gcp.native_cloud_connection = AAZObjectType(
                serialized_name="nativeCloudConnection",
            )

            native_cloud_connection = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "CspmMonitorGcp").native_cloud_connection
            native_cloud_connection.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            native_cloud_connection.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            disc_defender_cspm_aws = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws")
            disc_defender_cspm_aws.ciem = AAZObjectType()
            disc_defender_cspm_aws.data_sensitivity_discovery = AAZObjectType(
                serialized_name="dataSensitivityDiscovery",
            )
            disc_defender_cspm_aws.databases_dspm = AAZObjectType(
                serialized_name="databasesDspm",
            )
            disc_defender_cspm_aws.mdc_containers_agentless_discovery_k8s = AAZObjectType(
                serialized_name="mdcContainersAgentlessDiscoveryK8s",
            )
            disc_defender_cspm_aws.mdc_containers_image_assessment = AAZObjectType(
                serialized_name="mdcContainersImageAssessment",
            )
            disc_defender_cspm_aws.vm_scanners = AAZObjectType(
                serialized_name="vmScanners",
            )

            ciem = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").ciem
            ciem.ciem_discovery = AAZObjectType(
                serialized_name="ciemDiscovery",
            )
            ciem.ciem_oidc = AAZObjectType(
                serialized_name="ciemOidc",
            )

            ciem_discovery = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").ciem.ciem_discovery
            ciem_discovery.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            ciem_oidc = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").ciem.ciem_oidc
            ciem_oidc.azure_active_directory_app_name = AAZStrType(
                serialized_name="azureActiveDirectoryAppName",
            )
            ciem_oidc.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            data_sensitivity_discovery = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").data_sensitivity_discovery
            data_sensitivity_discovery.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            data_sensitivity_discovery.enabled = AAZBoolType()

            databases_dspm = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").databases_dspm
            databases_dspm.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            databases_dspm.enabled = AAZBoolType()

            mdc_containers_agentless_discovery_k8s = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").mdc_containers_agentless_discovery_k8s
            mdc_containers_agentless_discovery_k8s.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            mdc_containers_agentless_discovery_k8s.enabled = AAZBoolType()

            mdc_containers_image_assessment = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").mdc_containers_image_assessment
            mdc_containers_image_assessment.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            mdc_containers_image_assessment.enabled = AAZBoolType()

            vm_scanners = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").vm_scanners
            vm_scanners.configuration = AAZObjectType()
            vm_scanners.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").vm_scanners.configuration
            configuration.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            configuration.exclusion_tags = AAZDictType(
                serialized_name="exclusionTags",
            )
            configuration.scanning_mode = AAZStrType(
                serialized_name="scanningMode",
            )

            exclusion_tags = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmAws").vm_scanners.configuration.exclusion_tags
            exclusion_tags.Element = AAZStrType()

            disc_defender_cspm_gcp = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp")
            disc_defender_cspm_gcp.ciem_discovery = AAZObjectType(
                serialized_name="ciemDiscovery",
            )
            disc_defender_cspm_gcp.data_sensitivity_discovery = AAZObjectType(
                serialized_name="dataSensitivityDiscovery",
            )
            disc_defender_cspm_gcp.mdc_containers_agentless_discovery_k8s = AAZObjectType(
                serialized_name="mdcContainersAgentlessDiscoveryK8s",
            )
            disc_defender_cspm_gcp.mdc_containers_image_assessment = AAZObjectType(
                serialized_name="mdcContainersImageAssessment",
            )
            disc_defender_cspm_gcp.vm_scanners = AAZObjectType(
                serialized_name="vmScanners",
            )

            ciem_discovery = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp").ciem_discovery
            ciem_discovery.azure_active_directory_app_name = AAZStrType(
                serialized_name="azureActiveDirectoryAppName",
            )
            ciem_discovery.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            ciem_discovery.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            data_sensitivity_discovery = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp").data_sensitivity_discovery
            data_sensitivity_discovery.enabled = AAZBoolType()
            data_sensitivity_discovery.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            data_sensitivity_discovery.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            mdc_containers_agentless_discovery_k8s = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp").mdc_containers_agentless_discovery_k8s
            mdc_containers_agentless_discovery_k8s.enabled = AAZBoolType()
            mdc_containers_agentless_discovery_k8s.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            mdc_containers_agentless_discovery_k8s.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            mdc_containers_image_assessment = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp").mdc_containers_image_assessment
            mdc_containers_image_assessment.enabled = AAZBoolType()
            mdc_containers_image_assessment.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            mdc_containers_image_assessment.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            vm_scanners = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp").vm_scanners
            vm_scanners.configuration = AAZObjectType()
            vm_scanners.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp").vm_scanners.configuration
            configuration.exclusion_tags = AAZDictType(
                serialized_name="exclusionTags",
            )
            configuration.scanning_mode = AAZStrType(
                serialized_name="scanningMode",
            )

            exclusion_tags = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderCspmGcp").vm_scanners.configuration.exclusion_tags
            exclusion_tags.Element = AAZStrType()

            disc_defender_for_containers_aws = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws")
            disc_defender_for_containers_aws.auto_provisioning = AAZBoolType(
                serialized_name="autoProvisioning",
            )
            disc_defender_for_containers_aws.cloud_watch_to_kinesis = AAZObjectType(
                serialized_name="cloudWatchToKinesis",
            )
            disc_defender_for_containers_aws.container_vulnerability_assessment = AAZObjectType(
                serialized_name="containerVulnerabilityAssessment",
            )
            disc_defender_for_containers_aws.container_vulnerability_assessment_task = AAZObjectType(
                serialized_name="containerVulnerabilityAssessmentTask",
            )
            disc_defender_for_containers_aws.enable_container_vulnerability_assessment = AAZBoolType(
                serialized_name="enableContainerVulnerabilityAssessment",
            )
            disc_defender_for_containers_aws.kinesis_to_s3 = AAZObjectType(
                serialized_name="kinesisToS3",
            )
            disc_defender_for_containers_aws.kube_audit_retention_time = AAZIntType(
                serialized_name="kubeAuditRetentionTime",
            )
            disc_defender_for_containers_aws.kubernetes_scuba_reader = AAZObjectType(
                serialized_name="kubernetesScubaReader",
            )
            disc_defender_for_containers_aws.kubernetes_service = AAZObjectType(
                serialized_name="kubernetesService",
            )
            disc_defender_for_containers_aws.mdc_containers_agentless_discovery_k8s = AAZObjectType(
                serialized_name="mdcContainersAgentlessDiscoveryK8s",
            )
            disc_defender_for_containers_aws.mdc_containers_image_assessment = AAZObjectType(
                serialized_name="mdcContainersImageAssessment",
            )
            disc_defender_for_containers_aws.scuba_external_id = AAZStrType(
                serialized_name="scubaExternalId",
            )

            cloud_watch_to_kinesis = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").cloud_watch_to_kinesis
            cloud_watch_to_kinesis.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            container_vulnerability_assessment = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").container_vulnerability_assessment
            container_vulnerability_assessment.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            container_vulnerability_assessment_task = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").container_vulnerability_assessment_task
            container_vulnerability_assessment_task.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            kinesis_to_s3 = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").kinesis_to_s3
            kinesis_to_s3.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            kubernetes_scuba_reader = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").kubernetes_scuba_reader
            kubernetes_scuba_reader.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            kubernetes_service = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").kubernetes_service
            kubernetes_service.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            mdc_containers_agentless_discovery_k8s = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").mdc_containers_agentless_discovery_k8s
            mdc_containers_agentless_discovery_k8s.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            mdc_containers_agentless_discovery_k8s.enabled = AAZBoolType()

            mdc_containers_image_assessment = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersAws").mdc_containers_image_assessment
            mdc_containers_image_assessment.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            mdc_containers_image_assessment.enabled = AAZBoolType()

            disc_defender_for_containers_gcp = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersGcp")
            disc_defender_for_containers_gcp.audit_logs_auto_provisioning_flag = AAZBoolType(
                serialized_name="auditLogsAutoProvisioningFlag",
            )
            disc_defender_for_containers_gcp.data_pipeline_native_cloud_connection = AAZObjectType(
                serialized_name="dataPipelineNativeCloudConnection",
            )
            disc_defender_for_containers_gcp.defender_agent_auto_provisioning_flag = AAZBoolType(
                serialized_name="defenderAgentAutoProvisioningFlag",
            )
            disc_defender_for_containers_gcp.mdc_containers_agentless_discovery_k8s = AAZObjectType(
                serialized_name="mdcContainersAgentlessDiscoveryK8s",
            )
            disc_defender_for_containers_gcp.mdc_containers_image_assessment = AAZObjectType(
                serialized_name="mdcContainersImageAssessment",
            )
            disc_defender_for_containers_gcp.native_cloud_connection = AAZObjectType(
                serialized_name="nativeCloudConnection",
            )
            disc_defender_for_containers_gcp.policy_agent_auto_provisioning_flag = AAZBoolType(
                serialized_name="policyAgentAutoProvisioningFlag",
            )

            data_pipeline_native_cloud_connection = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersGcp").data_pipeline_native_cloud_connection
            data_pipeline_native_cloud_connection.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            data_pipeline_native_cloud_connection.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            mdc_containers_agentless_discovery_k8s = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersGcp").mdc_containers_agentless_discovery_k8s
            mdc_containers_agentless_discovery_k8s.enabled = AAZBoolType()
            mdc_containers_agentless_discovery_k8s.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            mdc_containers_agentless_discovery_k8s.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            mdc_containers_image_assessment = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersGcp").mdc_containers_image_assessment
            mdc_containers_image_assessment.enabled = AAZBoolType()
            mdc_containers_image_assessment.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            mdc_containers_image_assessment.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            native_cloud_connection = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForContainersGcp").native_cloud_connection
            native_cloud_connection.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            native_cloud_connection.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            disc_defender_for_databases_aws = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesAws")
            disc_defender_for_databases_aws.arc_auto_provisioning = AAZObjectType(
                serialized_name="arcAutoProvisioning",
            )
            disc_defender_for_databases_aws.databases_dspm = AAZObjectType(
                serialized_name="databasesDspm",
            )
            disc_defender_for_databases_aws.rds = AAZObjectType()

            arc_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesAws").arc_auto_provisioning
            arc_auto_provisioning.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            arc_auto_provisioning.configuration = AAZObjectType()
            arc_auto_provisioning.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesAws").arc_auto_provisioning.configuration
            configuration.private_link_scope = AAZStrType(
                serialized_name="privateLinkScope",
            )
            configuration.proxy = AAZStrType()

            databases_dspm = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesAws").databases_dspm
            databases_dspm.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            databases_dspm.enabled = AAZBoolType()

            rds = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesAws").rds
            rds.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            rds.enabled = AAZBoolType()

            disc_defender_for_databases_gcp = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesGcp")
            disc_defender_for_databases_gcp.arc_auto_provisioning = AAZObjectType(
                serialized_name="arcAutoProvisioning",
            )
            disc_defender_for_databases_gcp.defender_for_databases_arc_auto_provisioning = AAZObjectType(
                serialized_name="defenderForDatabasesArcAutoProvisioning",
            )

            arc_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesGcp").arc_auto_provisioning
            arc_auto_provisioning.configuration = AAZObjectType()
            arc_auto_provisioning.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesGcp").arc_auto_provisioning.configuration
            configuration.private_link_scope = AAZStrType(
                serialized_name="privateLinkScope",
            )
            configuration.proxy = AAZStrType()

            defender_for_databases_arc_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForDatabasesGcp").defender_for_databases_arc_auto_provisioning
            defender_for_databases_arc_auto_provisioning.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            defender_for_databases_arc_auto_provisioning.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            disc_defender_for_servers_aws = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws")
            disc_defender_for_servers_aws.arc_auto_provisioning = AAZObjectType(
                serialized_name="arcAutoProvisioning",
            )
            disc_defender_for_servers_aws.defender_for_servers = AAZObjectType(
                serialized_name="defenderForServers",
            )
            disc_defender_for_servers_aws.mde_auto_provisioning = AAZObjectType(
                serialized_name="mdeAutoProvisioning",
            )
            disc_defender_for_servers_aws.sub_plan = AAZObjectType(
                serialized_name="subPlan",
            )
            disc_defender_for_servers_aws.va_auto_provisioning = AAZObjectType(
                serialized_name="vaAutoProvisioning",
            )
            disc_defender_for_servers_aws.vm_scanners = AAZObjectType(
                serialized_name="vmScanners",
            )

            arc_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").arc_auto_provisioning
            arc_auto_provisioning.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            arc_auto_provisioning.configuration = AAZObjectType()
            arc_auto_provisioning.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").arc_auto_provisioning.configuration
            configuration.private_link_scope = AAZStrType(
                serialized_name="privateLinkScope",
            )
            configuration.proxy = AAZStrType()

            defender_for_servers = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").defender_for_servers
            defender_for_servers.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            mde_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").mde_auto_provisioning
            mde_auto_provisioning.configuration = AAZObjectType()
            mde_auto_provisioning.enabled = AAZBoolType()

            sub_plan = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").sub_plan
            sub_plan.type = AAZStrType()

            va_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").va_auto_provisioning
            va_auto_provisioning.configuration = AAZObjectType()
            va_auto_provisioning.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").va_auto_provisioning.configuration
            configuration.type = AAZStrType()

            vm_scanners = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").vm_scanners
            vm_scanners.configuration = AAZObjectType()
            vm_scanners.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").vm_scanners.configuration
            configuration.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )
            configuration.exclusion_tags = AAZDictType(
                serialized_name="exclusionTags",
            )
            configuration.scanning_mode = AAZStrType(
                serialized_name="scanningMode",
            )

            exclusion_tags = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersAws").vm_scanners.configuration.exclusion_tags
            exclusion_tags.Element = AAZStrType()

            disc_defender_for_servers_gcp = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp")
            disc_defender_for_servers_gcp.arc_auto_provisioning = AAZObjectType(
                serialized_name="arcAutoProvisioning",
            )
            disc_defender_for_servers_gcp.defender_for_servers = AAZObjectType(
                serialized_name="defenderForServers",
            )
            disc_defender_for_servers_gcp.mde_auto_provisioning = AAZObjectType(
                serialized_name="mdeAutoProvisioning",
            )
            disc_defender_for_servers_gcp.sub_plan = AAZObjectType(
                serialized_name="subPlan",
            )
            disc_defender_for_servers_gcp.va_auto_provisioning = AAZObjectType(
                serialized_name="vaAutoProvisioning",
            )
            disc_defender_for_servers_gcp.vm_scanners = AAZObjectType(
                serialized_name="vmScanners",
            )

            arc_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").arc_auto_provisioning
            arc_auto_provisioning.configuration = AAZObjectType()
            arc_auto_provisioning.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").arc_auto_provisioning.configuration
            configuration.private_link_scope = AAZStrType(
                serialized_name="privateLinkScope",
            )
            configuration.proxy = AAZStrType()

            defender_for_servers = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").defender_for_servers
            defender_for_servers.service_account_email_address = AAZStrType(
                serialized_name="serviceAccountEmailAddress",
            )
            defender_for_servers.workload_identity_provider_id = AAZStrType(
                serialized_name="workloadIdentityProviderId",
            )

            mde_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").mde_auto_provisioning
            mde_auto_provisioning.configuration = AAZObjectType()
            mde_auto_provisioning.enabled = AAZBoolType()

            sub_plan = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").sub_plan
            sub_plan.type = AAZStrType()

            va_auto_provisioning = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").va_auto_provisioning
            va_auto_provisioning.configuration = AAZObjectType()
            va_auto_provisioning.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").va_auto_provisioning.configuration
            configuration.type = AAZStrType()

            vm_scanners = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").vm_scanners
            vm_scanners.configuration = AAZObjectType()
            vm_scanners.enabled = AAZBoolType()

            configuration = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").vm_scanners.configuration
            configuration.exclusion_tags = AAZDictType(
                serialized_name="exclusionTags",
            )
            configuration.scanning_mode = AAZStrType(
                serialized_name="scanningMode",
            )

            exclusion_tags = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "DefenderForServersGcp").vm_scanners.configuration.exclusion_tags
            exclusion_tags.Element = AAZStrType()

            disc_information_protection_aws = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "InformationProtectionAws")
            disc_information_protection_aws.information_protection = AAZObjectType(
                serialized_name="informationProtection",
            )

            information_protection = cls._schema_on_200.properties.offerings.Element.discriminate_by("offering_type", "InformationProtectionAws").information_protection
            information_protection.cloud_role_arn = AAZStrType(
                serialized_name="cloudRoleArn",
            )

            system_data = cls._schema_on_200.system_data
            system_data.created_at = AAZStrType(
                serialized_name="createdAt",
            )
            system_data.created_by = AAZStrType(
                serialized_name="createdBy",
            )
            system_data.created_by_type = AAZStrType(
                serialized_name="createdByType",
            )
            system_data.last_modified_at = AAZStrType(
                serialized_name="lastModifiedAt",
            )
            system_data.last_modified_by = AAZStrType(
                serialized_name="lastModifiedBy",
            )
            system_data.last_modified_by_type = AAZStrType(
                serialized_name="lastModifiedByType",
            )

            tags = cls._schema_on_200.tags
            tags.Element = AAZStrType()

            return cls._schema_on_200


class _ShowHelper:
    """Helper class for Show"""


__all__ = ["Show"]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from msrestazure.tools import is_valid_resource_id, resource_id
from knack.prompting import prompt_pass
from azure.cli.core.azclierror import (
    RequiredArgumentMissingError,
    AzureResponseError,
    HTTPError
)
from azure.cli.core.commands import LongRunningOperation
from azure.cli.core.util import (
    sdk_no_wait
)

from azure.mgmt.sqlvirtualmachine.models import (
    WsfcDomainProfile,
    SqlVirtualMachineGroup,
    PrivateIPAddress,
    LoadBalancerConfiguration,
    AvailabilityGroupListener,
    WsfcDomainCredentials,
    AutoPatchingSettings,
    AutoBackupSettings,
    KeyVaultCredentialSettings,
    SqlConnectivityUpdateSettings,
    SqlWorkloadTypeUpdateSettings,
    AdditionalFeaturesServerConfigurations,
    ServerConfigurationsManagementSettings,
    Schedule,
    AssessmentSettings,
    SqlVirtualMachine,
    AADAuthenticationSettings
)

from ._util import (
    does_custom_log_exist,
    create_custom_table,
    validate_dcr,
    does_name_exist,
    create_ama_and_dcra
)

from azure.cli.command_modules.sqlvm._template_builder import (
    build_dcr_resource,
    build_dce_resource
)
import re

# from azure.mgmt.resource import ResourceManagementClient


def sqlvm_list(
        client,
        resource_group_name=None):
    '''
    Lists all SQL virtual machines in a resource group or subscription.
    '''
    if resource_group_name:
        # List all sql vms  in the resource group
        return client.list_by_resource_group(
            resource_group_name=resource_group_name)

    # List all sql vms in the subscription
    return client.list()


def sqlvm_group_list(
        client,
        resource_group_name=None):
    '''
    Lists all SQL virtual machine groups in a resource group or subscription.
    '''
    if resource_group_name:
        # List all sql vm groups in the resource group
        return client.list_by_resource_group(
            resource_group_name=resource_group_name)

    # List all sql vm groups in the subscription
    return client.list()


# pylint: disable= line-too-long, too-many-arguments
def sqlvm_group_create(
        client,
        cmd,
        sql_virtual_machine_group_name,
        resource_group_name,
        sql_image_offer,
        sql_image_sku,
        domain_fqdn,
        cluster_operator_account,
        sql_service_account,
        storage_account_url,
        cluster_subnet_type="SingleSubnet",
        storage_account_key=None,
        location=None,
        cluster_bootstrap_account=None,
        file_share_witness_path=None,
        ou_path=None,
        tags=None):
    '''
    Creates a SQL virtual machine group.
    '''
    tags = tags or {}

    if not storage_account_key:
        storage_account_key = prompt_pass('Storage Key: ', confirm=True)

    # Create the windows server failover cluster domain profile object.
    wsfc_domain_profile_object = WsfcDomainProfile(
        domain_fqdn=domain_fqdn,
        ou_path=ou_path,
        cluster_bootstrap_account=cluster_bootstrap_account,
        cluster_operator_account=cluster_operator_account,
        sql_service_account=sql_service_account,
        file_share_witness_path=file_share_witness_path,
        storage_account_url=storage_account_url,
        storage_account_primary_key=storage_account_key,
        cluster_subnet_type=cluster_subnet_type)

    sqlvm_group_object = SqlVirtualMachineGroup(
        sql_image_offer=sql_image_offer,
        sql_image_sku=sql_image_sku,
        wsfc_domain_profile=wsfc_domain_profile_object,
        location=location,
        tags=tags)

    # Since it's a running operation, we will do the put and then the get to
    # display the instance.
    LongRunningOperation(
        cmd.cli_ctx)(
        sdk_no_wait(
            False,
            client.begin_create_or_update,
            resource_group_name,
            sql_virtual_machine_group_name,
            sqlvm_group_object))

    return client.get(resource_group_name, sql_virtual_machine_group_name)


# pylint: disable=line-too-long, too-many-arguments
def sqlvm_group_update(
        instance,
        domain_fqdn=None,
        cluster_operator_account=None,
        sql_service_account=None,
        storage_account_url=None,
        storage_account_key=None,
        cluster_bootstrap_account=None,
        file_share_witness_path=None,
        ou_path=None,
        tags=None,
        cluster_subnet_type=None):
    '''
    Updates a SQL virtual machine group.
    '''
    if domain_fqdn is not None:
        instance.wsfc_domain_profile.domain_fqdn = domain_fqdn
    if cluster_operator_account is not None:
        instance.wsfc_domain_profile.cluster_operator_account = cluster_operator_account
    if cluster_bootstrap_account is not None:
        instance.wsfc_domain_profile.cluster_bootstrap_account = cluster_bootstrap_account
    if sql_service_account is not None:
        instance.wsfc_domain_profile.sql_service_account = sql_service_account
    if storage_account_url is not None:
        instance.wsfc_domain_profile.storage_account_url = storage_account_url
    if storage_account_key is not None:
        instance.wsfc_domain_profile.storage_account_primary_key = storage_account_key
    if storage_account_url and not storage_account_key:
        instance.wsfc_domain_profile.storage_account_primary_key = prompt_pass(
            'Storage Key: ', confirm=True)
    if file_share_witness_path is not None:
        instance.wsfc_domain_profile.file_share_witness_path = file_share_witness_path
    if ou_path is not None:
        instance.wsfc_domain_profile.ou_path = ou_path
    if cluster_subnet_type is not None:
        instance.wsfc_domain_profile.cluster_subnet_type = cluster_subnet_type
    if tags is not None:
        instance.tags = tags

    return instance


# pylint: disable=line-too-long, too-many-boolean-expressions, too-many-arguments
def sqlvm_aglistener_create(client, cmd, availability_group_listener_name, sql_virtual_machine_group_name,
                            resource_group_name, availability_group_name, ip_address, subnet_resource_id,
                            load_balancer_resource_id, probe_port, sql_virtual_machine_instances, port=1433,
                            public_ip_address_resource_id=None, vnet_name=None):  # pylint: disable=unused-argument
    '''
    Creates an availability group listener
    '''
    # Create the private ip address
    private_ip_object = PrivateIPAddress(ip_address=ip_address,
                                         subnet_resource_id=subnet_resource_id
                                         if is_valid_resource_id(subnet_resource_id) else None)

    # Create the load balancer configurations
    load_balancer_object = LoadBalancerConfiguration(
        private_ip_address=private_ip_object,
        public_ip_address_resource_id=public_ip_address_resource_id,
        load_balancer_resource_id=load_balancer_resource_id,
        probe_port=probe_port,
        sql_virtual_machine_instances=sql_virtual_machine_instances)

    # Create the availability group listener object
    ag_listener_object = AvailabilityGroupListener(
        availability_group_name=availability_group_name,
        load_balancer_configurations=[load_balancer_object],
        port=port)

    LongRunningOperation(
        cmd.cli_ctx)(
        sdk_no_wait(
            False,
            client.begin_create_or_update,
            resource_group_name,
            sql_virtual_machine_group_name,
            availability_group_listener_name,
            ag_listener_object))

    return client.get(
        resource_group_name,
        sql_virtual_machine_group_name,
        availability_group_listener_name)


def aglistener_update(instance, sql_virtual_machine_instances=None):
    '''
    Updates an availability group listener
    '''
    # Get the list of all current machines in the ag listener
    if sql_virtual_machine_instances:
        instance.load_balancer_configurations[0].sql_virtual_machine_instances = sql_virtual_machine_instances

    return instance


# pylint: disable=too-many-arguments, too-many-locals, line-too-long, too-many-boolean-expressions
def sqlvm_create(
        client,
        cmd,
        sql_virtual_machine_name,
        resource_group_name,
        sql_server_license_type=None,
        location=None,
        sql_image_sku=None,
        enable_auto_patching=None,
        sql_management_mode="LightWeight",
        least_privilege_mode=None,
        day_of_week=None,
        maintenance_window_starting_hour=None,
        maintenance_window_duration=None,
        enable_auto_backup=None,
        enable_encryption=False,
        retention_period=None,
        storage_account_url=None,
        storage_access_key=None,
        backup_password=None,
        backup_system_dbs=False,
        backup_schedule_type=None,
        full_backup_frequency=None,
        full_backup_start_time=None,
        full_backup_window_hours=None,
        log_backup_frequency=None,
        enable_key_vault_credential=None,
        credential_name=None,
        azure_key_vault_url=None,
        service_principal_name=None,
        service_principal_secret=None,
        connectivity_type=None,
        port=None,
        sql_auth_update_username=None,
        sql_auth_update_password=None,
        sql_workload_type=None,
        enable_r_services=None,
        tags=None,
        sql_image_offer=None):
    '''
    Creates a SQL virtual machine.
    '''
    from azure.cli.core.commands.client_factory import get_subscription_id

    subscription_id = get_subscription_id(cmd.cli_ctx)

    virtual_machine_resource_id = resource_id(
        subscription=subscription_id,
        resource_group=resource_group_name,
        namespace='Microsoft.Compute',
        type='virtualMachines',
        name=sql_virtual_machine_name)

    tags = tags or {}

    # If customer has provided any auto_patching settings, enabling plugin
    # should be True
    if (day_of_week or maintenance_window_duration or maintenance_window_starting_hour):
        enable_auto_patching = True

    auto_patching_object = AutoPatchingSettings(
        enable=enable_auto_patching,
        day_of_week=day_of_week,
        maintenance_window_starting_hour=maintenance_window_starting_hour,
        maintenance_window_duration=maintenance_window_duration)

    # If customer has provided any auto_backup settings, enabling plugin
    # should be True
    if (enable_encryption or retention_period or storage_account_url or storage_access_key or backup_password or
            backup_system_dbs or backup_schedule_type or full_backup_frequency or full_backup_start_time or
            full_backup_window_hours or log_backup_frequency):
        enable_auto_backup = True
        if not storage_access_key:
            storage_access_key = prompt_pass('Storage Key: ', confirm=True)
        if enable_encryption and not backup_password:
            backup_password = prompt_pass('Backup Password: ', confirm=True)

    auto_backup_object = AutoBackupSettings(
        enable=enable_auto_backup,
        enable_encryption=enable_encryption if enable_auto_backup else None,
        retention_period=retention_period,
        storage_account_url=storage_account_url,
        storage_access_key=storage_access_key,
        password=backup_password,
        backup_system_dbs=backup_system_dbs if enable_auto_backup else None,
        backup_schedule_type=backup_schedule_type,
        full_backup_frequency=full_backup_frequency,
        full_backup_start_time=full_backup_start_time,
        full_backup_window_hours=full_backup_window_hours,
        log_backup_frequency=log_backup_frequency)

    # If customer has provided any key_vault_credential settings, enabling
    # plugin should be True
    if (credential_name or azure_key_vault_url or service_principal_name or service_principal_secret):
        enable_key_vault_credential = True
        if not service_principal_secret:
            service_principal_secret = prompt_pass(
                'Service Principal Secret: ', confirm=True)

    keyvault_object = KeyVaultCredentialSettings(
        enable=enable_key_vault_credential,
        credential_name=credential_name,
        azure_key_vault_url=azure_key_vault_url,
        service_principal_name=service_principal_name,
        service_principal_secret=service_principal_secret)

    connectivity_object = SqlConnectivityUpdateSettings(
        port=port,
        connectivity_type=connectivity_type,
        sql_auth_update_user_name=sql_auth_update_username,
        sql_auth_update_password=sql_auth_update_password)

    workload_type_object = SqlWorkloadTypeUpdateSettings(
        sql_workload_type=sql_workload_type)

    additional_features_object = AdditionalFeaturesServerConfigurations(
        is_r_services_enabled=enable_r_services)

    server_configuration_object = ServerConfigurationsManagementSettings(
        sql_connectivity_update_settings=connectivity_object,
        sql_workload_type_update_settings=workload_type_object,
        additional_features_server_configurations=additional_features_object)

    sqlvm_object = SqlVirtualMachine(
        location=location,
        virtual_machine_resource_id=virtual_machine_resource_id,
        sql_server_license_type=sql_server_license_type,
        least_privilege_mode=least_privilege_mode,
        sql_image_sku=sql_image_sku,
        sql_management=sql_management_mode,
        sql_image_offer=sql_image_offer,
        auto_patching_settings=auto_patching_object,
        auto_backup_settings=auto_backup_object,
        key_vault_credential_settings=keyvault_object,
        server_configurations_management_settings=server_configuration_object,
        tags=tags)

    # Since it's a running operation, we will do the put and then the get to
    # display the instance.
    LongRunningOperation(
        cmd.cli_ctx)(
        sdk_no_wait(
            False,
            client.begin_create_or_update,
            resource_group_name,
            sql_virtual_machine_name,
            sqlvm_object))

    return client.get(resource_group_name, sql_virtual_machine_name)


# pylint: disable=too-many-statements, line-too-long, too-many-boolean-expressions, unused-argument
def sqlvm_update(
        cmd,
        instance,
        sql_virtual_machine_name,
        resource_group_name,
        sql_server_license_type=None,
        sql_image_sku=None,
        least_privilege_mode=None,
        enable_auto_patching=None,
        day_of_week=None,
        maintenance_window_starting_hour=None,
        maintenance_window_duration=None,
        enable_auto_backup=None,
        enable_encryption=False,
        retention_period=None,
        storage_account_url=None,
        prompt=True,
        storage_access_key=None,
        backup_password=None,
        backup_system_dbs=False,
        backup_schedule_type=None,
        sql_management_mode=None,
        full_backup_frequency=None,
        full_backup_start_time=None,
        full_backup_window_hours=None,
        log_backup_frequency=None,
        enable_key_vault_credential=None,
        credential_name=None,
        azure_key_vault_url=None,
        service_principal_name=None,
        service_principal_secret=None,
        connectivity_type=None,
        port=None,
        sql_workload_type=None,
        enable_r_services=None,
        tags=None,
        enable_assessment=None,
        enable_assessment_schedule=None,
        assessment_weekly_interval=None,
        assessment_monthly_occurrence=None,
        assessment_day_of_week=None,
        assessment_start_time_local=None,
        workspace_name=None,
        workspace_rg=None,
        workspace_sub=None,
        agent_rg=None):
    '''
    Updates a SQL virtual machine.
    '''
    if tags is not None:
        instance.tags = tags
    if sql_server_license_type is not None:
        instance.sql_server_license_type = sql_server_license_type
    if sql_image_sku is not None:
        instance.sql_image_sku = sql_image_sku
    if sql_management_mode is not None:
        instance.sql_management = sql_management_mode
    if least_privilege_mode is not None:
        instance.least_privilege_mode = least_privilege_mode

    if (enable_auto_patching is not None or day_of_week is not None or maintenance_window_starting_hour is not None or maintenance_window_duration is not None):

        enable_auto_patching = enable_auto_patching if enable_auto_patching is False else True
        instance.auto_patching_settings = AutoPatchingSettings(
            enable=enable_auto_patching,
            day_of_week=day_of_week,
            maintenance_window_starting_hour=maintenance_window_starting_hour,
            maintenance_window_duration=maintenance_window_duration)

    if (enable_auto_backup is not None or enable_encryption or retention_period is not None or storage_account_url is not None or
            storage_access_key is not None or backup_password is not None or backup_system_dbs or backup_schedule_type is not None or
            full_backup_frequency is not None or full_backup_start_time is not None or full_backup_window_hours is not None or
            log_backup_frequency is not None):

        enable_auto_backup = enable_auto_backup if enable_auto_backup is False else True
        if not storage_access_key:
            storage_access_key = prompt_pass('Storage Key: ', confirm=True)
        if enable_encryption and not backup_password:
            backup_password = prompt_pass('Backup Password: ', confirm=True)

        instance.auto_backup_settings = AutoBackupSettings(
            enable=enable_auto_backup,
            enable_encryption=enable_encryption if enable_auto_backup else None,
            retention_period=retention_period,
            storage_account_url=storage_account_url,
            storage_access_key=storage_access_key,
            password=backup_password,
            backup_system_dbs=backup_system_dbs if enable_auto_backup else None,
            backup_schedule_type=backup_schedule_type,
            full_backup_frequency=full_backup_frequency,
            full_backup_start_time=full_backup_start_time,
            full_backup_window_hours=full_backup_window_hours,
            log_backup_frequency=log_backup_frequency)

    if (enable_key_vault_credential is not None or credential_name is not None or azure_key_vault_url is not None or
            service_principal_name is not None or service_principal_secret is not None):

        enable_key_vault_credential = enable_key_vault_credential if enable_key_vault_credential is False else True
        if not service_principal_secret:
            service_principal_secret = prompt_pass(
                'Service Principal Secret: ', confirm=True)

        instance.key_vault_credential_settings = KeyVaultCredentialSettings(
            enable=enable_key_vault_credential,
            credential_name=credential_name,
            service_principal_name=service_principal_name,
            service_principal_secret=service_principal_secret,
            azure_key_vault_url=azure_key_vault_url)

    instance.server_configurations_management_settings = ServerConfigurationsManagementSettings()

    if (connectivity_type is not None or port is not None):
        instance.server_configurations_management_settings.sql_connectivity_update_settings = SqlConnectivityUpdateSettings(
            connectivity_type=connectivity_type, port=port)

    if sql_workload_type is not None:
        instance.server_configurations_management_settings.sql_workload_type_update_settings = SqlWorkloadTypeUpdateSettings(
            sql_workload_type=sql_workload_type)

    if enable_r_services is not None:
        instance.server_configurations_management_settings.additional_features_server_configurations = AdditionalFeaturesServerConfigurations(
            is_r_services_enabled=enable_r_services)

    # If none of the settings was modified, reset
    # server_configurations_management_settings to be null
    if (instance.server_configurations_management_settings.sql_connectivity_update_settings is None and
            instance.server_configurations_management_settings.sql_workload_type_update_settings is None and
            instance.server_configurations_management_settings.sql_storage_update_settings is None and
            instance.server_configurations_management_settings.additional_features_server_configurations is None):
        instance.server_configurations_management_settings = None

    set_assessment_properties(cmd,
                              instance,
                              enable_assessment,
                              enable_assessment_schedule,
                              assessment_weekly_interval,
                              assessment_monthly_occurrence,
                              assessment_day_of_week,
                              assessment_start_time_local,
                              resource_group_name,
                              sql_virtual_machine_name,
                              workspace_rg,
                              workspace_name,
                              workspace_sub,
                              agent_rg)

    return instance


# pylint: disable=unused-argument
def sqlvm_enable_azure_ad_auth(client, cmd, sql_virtual_machine_name, resource_group_name, msi_client_id=None, skip_client_validation=None):
    ''' Enable Azure AD authentication on a SQL virtual machine.

        :param cmd: The CLI command.
        :type cmd: AzCliCommand.
        :param instance: The Sql Virtual Machine instance.
        :type instance: SqlVirtualMachine.
        :param resource_group_name: The resource group name
        :type resource_group_name: str.
        :param msi_client_id: The clientId of the managed identity used in Azure AD authentication.
                              None means system-assigned managed identity
        :type: str.
        :param skip_client_validation: Whether to skip the client side validation. The server side validation always happens.
                                       This parameter is used in the validation and ignored here.
        :type: bool.

        :return: The updated Sql Virtual Machine instance.
        :rtype: SqlVirtualMachine.
    '''

    sqlvm_object = client.get(resource_group_name, sql_virtual_machine_name)

    if sqlvm_object.server_configurations_management_settings is None:
        sqlvm_object.server_configurations_management_settings = ServerConfigurationsManagementSettings()

    sqlvm_object.server_configurations_management_settings.azure_ad_authentication_settings = AADAuthenticationSettings(client_id=msi_client_id if msi_client_id else '')

    # Since it's a running operation, we will do the put and then the get to display the instance.
    LongRunningOperation(cmd.cli_ctx)(sdk_no_wait(False, client.begin_create_or_update,
                                                  resource_group_name, sql_virtual_machine_name, sqlvm_object))

    return client.get(resource_group_name, sql_virtual_machine_name)


# pylint: disable=unused-argument
def validate_azure_ad_auth(cmd, sql_virtual_machine_name, resource_group_name, msi_client_id=None):
    ''' Valida if Azure AD authentication is ready on a SQL virtual machine.
        The logic of validation is in the validator method. If the SQL virtual machine passes the validator,
        it means this SQL virtual machine is valid for Azure AD authentication

        :param cmd: The CLI command.
        :type cmd: AzCliCommand.
        :param resource_group_name: The resource group name
        :type resource_group_name: str.
        :param msi_client_id: The clientId of the managed identity used in Azure AD authentication.
                              None means system-assigned managed identity
        :type: str.

        :return: The updated Sql Virtual Machine instance.
        :rtype: SqlVirtualMachine.
    '''

    passing_validation_message = "Sql virtual machine {} is valid for Azure AD authentication.".format(sql_virtual_machine_name)

    return passing_validation_message


def sqlvm_add_to_group(
        client,
        cmd,
        sql_virtual_machine_name,
        resource_group_name,
        sql_virtual_machine_group_resource_id,
        sql_service_account_password=None,
        cluster_operator_account_password=None,
        cluster_bootstrap_account_password=None):
    '''
    Adds a SQL virtual machine to a group.
    '''

    sqlvm_object = client.get(resource_group_name, sql_virtual_machine_name)

    if not sql_service_account_password:
        sql_service_account_password = prompt_pass(
            'SQL Service account password: ', confirm=True)
    if not cluster_operator_account_password:
        cluster_operator_account_password = prompt_pass(
            'Cluster operator account password: ',
            confirm=True,
            help_string='Password to authenticate with the domain controller.')

    sqlvm_object.sql_virtual_machine_group_resource_id = sql_virtual_machine_group_resource_id

    sqlvm_object.wsfc_domain_credentials = WsfcDomainCredentials(
        cluster_bootstrap_account_password=cluster_bootstrap_account_password,
        cluster_operator_account_password=cluster_operator_account_password,
        sql_service_account_password=sql_service_account_password)

    # Since it's a running operation, we will do the put and then the get to
    # display the instance.
    LongRunningOperation(
        cmd.cli_ctx)(
        sdk_no_wait(
            False,
            client.begin_create_or_update,
            resource_group_name,
            sql_virtual_machine_name,
            sqlvm_object))

    return client.get(resource_group_name, sql_virtual_machine_name)


def sqlvm_remove_from_group(
        client,
        cmd,
        sql_virtual_machine_name,
        resource_group_name):
    '''
    Removes a SQL virtual machine from a group.
    '''

    sqlvm_object = client.get(resource_group_name, sql_virtual_machine_name)

    sqlvm_object.sql_virtual_machine_group_resource_id = None
    sqlvm_object.wsfc_domain_credentials = None

    # Since it's a running operation, we will do the put and then the get to
    # display the instance.
    LongRunningOperation(
        cmd.cli_ctx)(
        sdk_no_wait(
            False,
            client.begin_create_or_update,
            resource_group_name,
            sql_virtual_machine_name,
            sqlvm_object))

    return client.get(resource_group_name, sql_virtual_machine_name)


# region Helpers for custom commands
# pylint: disable=too-many-branches
def set_assessment_properties(
        cmd,
        instance,
        enable_assessment,
        enable_assessment_schedule,
        assessment_weekly_interval,
        assessment_monthly_occurrence,
        assessment_day_of_week,
        assessment_start_time_local,
        resource_group_name,
        sql_virtual_machine_name,
        workspace_rg,
        workspace_name,
        workspace_sub,
        agent_rg):
    '''
    Set assessment properties to be sent in sql vm update
    '''

    from azure.cli.core.commands.client_factory import get_subscription_id
    from azure.cli.core.util import random_string, send_raw_request
    from azure.cli.command_modules.vm._vm_utils import ArmTemplateBuilder20190401
    from azure.cli.core.profiles import ResourceType
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    from ._assessment_data_source import table_name
    from itertools import count

    # If assessment.schedule settings are provided but enable schedule is
    # skipped, then ensure schedule is enabled
    if (enable_assessment_schedule is None and (
            assessment_weekly_interval is not None or assessment_monthly_occurrence or assessment_day_of_week or assessment_start_time_local)):
        enable_assessment_schedule = True

    # If assessment schedule is enabled but enable assessment is skipped, then
    # ensure assessment is enabled
    if (enable_assessment_schedule is not None and enable_assessment is None):
        enable_assessment = True

    if enable_assessment is not None:
        instance.assessment_settings = AssessmentSettings()
        instance.assessment_settings.enable = enable_assessment

        if enable_assessment_schedule is not None:
            instance.assessment_settings.schedule = Schedule()
            instance.assessment_settings.schedule.enable = enable_assessment_schedule

            if enable_assessment_schedule:
                instance.assessment_settings.schedule.weekly_interval = assessment_weekly_interval
                instance.assessment_settings.schedule.monthly_occurrence = assessment_monthly_occurrence
                instance.assessment_settings.schedule.day_of_week = assessment_day_of_week
                instance.assessment_settings.schedule.start_time = assessment_start_time_local

    # Validate and deploy pre-requisites if necessary
    # 1. Log Analytics extension for given workspace
    # 2. Custom log definition on workspace
    if enable_assessment:
        workspace_id = None
        curr_subscription = get_subscription_id(cmd.cli_ctx)
        # Raise error if workspace arguments not provided by user
        if workspace_name is None or workspace_rg is None:
            raise RequiredArgumentMissingError(
                'Assessment requires a Log Analytics workspace and Log Analytics extension on VM - '
                'workspace name and workspace resource group must be specified to deploy pre-requisites.')
        if agent_rg is None:
            agent_rg = resource_group_name
            # raise Warning -
            # raise RequiredArgumentMissingError(
            # 'Assessment requires a Resource Group to deploy the AMA Agent resources- '

        if workspace_sub is None:
            # raise warning => --workspace-sub not provided. Using current
            # subscription to find LA WS
            workspace_sub = curr_subscription
        api_version = "2021-12-01-preview"
        la_url = f"https://management.azure.com/subscriptions/{workspace_sub}/resourcegroups/{workspace_rg}/providers/Microsoft.OperationalInsights/workspaces/{workspace_name}?api-version={api_version}"

        try:
            la_response = send_raw_request(
                cmd.cli_ctx, method="GET", url=la_url)
        except Exception as e:
            raise AzureResponseError(
                f'Could not connect to the LA workspace - Error {e}.'
                ' If the workspace is not in the same subscription as the VM, use the --workspace-sub parameter')

        la_response = la_response.json()
        workspace_id = la_response['properties']['customerId']
        workspace_loc = la_response['location']

        workspace_res_id = la_response['id']

        # Validate the agent_rg -> Check if DCR + DCE exist already
        ama_sub = curr_subscription
        url = f"https://management.azure.com/subscriptions/{ama_sub}/resourceGroups/{agent_rg}/providers/Microsoft.Insights/dataCollectionRules?api-version=2022-06-01"
        try:
            dcr_response = send_raw_request(cmd.cli_ctx, method="GET", url=url)
        except HTTPError as e:
            raise AzureResponseError(
                f'An Http Error occured: {e}'
                'could not connect to the provided agent resource group {agent_rg}. Ensure the resource in the same subscription as {ama_sub}')

        # response contains list of dcr's
        dcr_response = dcr_response.json()
        dcr_list = dcr_response['value']
        dcr_found = False

        # get list of all dcr names found
        dcr_name_list = []

        for dcr in dcr_list:
            # Validate each dcr. Validation passes = reuse dcr

            # Fully qualified resource url that can be used
            dcr_id = dcr['id']

            # Define the regex pattern for extracting the required fields
            dcr_pattern = r'/subscriptions/([^/]+)/resourceGroups/([^/]+)/.*?/dataCollectionRules/([^/]+)'

            # Search the id in dcr_List for the fields
            dcr_match = re.search(dcr_pattern, dcr_id)

            if dcr_match:
                dcr_subId = dcr_match.group(1)
                dcr_rg = dcr_match.group(2)
                dcr_name = dcr_match.group(3)

            dcr_name_list.append(dcr_name)

            # Validate DCR Name with regex before continuing
            dcr_name_pattern = re.compile(
                r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}_[a-z0-9]+_DCR_\d+$",
                re.IGNORECASE)

            if dcr_name_pattern.match(dcr_name):
                url = f"https://management.azure.com/subscriptions/{dcr_subId}/resourceGroups/{dcr_rg}/providers/Microsoft.Insights/dataCollectionRules/{dcr_name}?api-version=2022-06-01"

                try:
                    dcr_response = send_raw_request(
                        cmd.cli_ctx, method="GET", url=url)
                except HTTPError:
                    # continue as we couldn't connect to DCR. If all
                    # connections fail we create new anyway
                    continue
            else:
                # If DCR Name doesn't match then skip to check next DCR
                continue
            # Validate dcr response
            dcr_response = dcr_response.json()

            dcr_location = dcr_response['location']
            dce_endpoint_id = dcr_response['properties']['dataCollectionEndpointId']
            dcr_source_filePattern = dcr_response['properties']['dataSources']['logFiles'][0]['filePatterns'][0]
            dcr_custom_log = dcr_response['properties']['dataFlows'][0]['outputStream']
            # Custom-SqlAssessment_CL
            dcr_la_id = dcr_response['properties']['destinations']['logAnalytics'][0]['workspaceId']
            # CustomerId is the workspace Id GUID. ResourceId is full qualified resource path
            # dcr_la_name = dcr_response['properties']['destinations']['logAnalytics'][0]['name']
            # workspace name is arbitrary name given by DCR resource metadata

            dcr_found = validate_dcr(
                cmd,
                dcr_location,
                workspace_loc,
                dcr_source_filePattern,
                dcr_custom_log,
                dcr_la_id,
                workspace_id,
                dce_endpoint_id)
            if dcr_found:
                validated_dcr_res_id = dcr_id
                break

            # Match all the stuff
            # collect a list of all dcr names found in this rg and ensure no
            # collision in new create name

            # Validate_DCR():
            # dcr follows naming convention
            # Sample DCR NAME => 0009fc4d-e310-4e40-8e63-c48a23e9cdc1_eastus_DCR_1
            # dcr location = la workspace location as they must be in same

        # New Custom table deployment workflow:
        # Check if old table exists - if yes - run POST command.
        # If does not exist add to the deployment template
        log_exists = does_custom_log_exist(
            cmd, workspace_name, workspace_rg, workspace_sub)

        # Check if log exists. V1 log checked with above. V2 log checked with
        # GET https://management.azure.com/subscriptions/{subscriptionId}/resourcegroups/{resourceGroupName}/providers/Microsoft.OperationalInsights/workspaces/{workspaceName}/tables/{tableName}?api-version=2021-12-01-preview
        # **This may be redundant if above checks v2 table by default **
        custom_table_url = f"https://management.azure.com/subscriptions/{curr_subscription}/resourceGroups/{workspace_rg}/providers/Microsoft.OperationalInsights/workspaces/{workspace_name}/tables/{table_name}?api-version=2021-12-01-preview"

        try:
            response = send_raw_request(
                cmd.cli_ctx, method="GET", url=custom_table_url)
            if response.status_code == 200:
                log_exists = True
            elif response.status_code == 404:
                log_exists = log_exists or False
            else:
                return False
        except HTTPError:
            log_exists = log_exists or False

        if log_exists:
            # Run a POST on the existing table for migration
            # POST
            # https://management.azure.com/subscriptions/{subscriptionId}/resourcegroups/{resourceGroupName}/providers/Microsoft.OperationalInsights/workspaces/{workspaceName}/tables/{tableName}/migrate?api-version=2021-12-01-preview
            try:
                # Does a GET on the dce to ensure no http errors - suffices
                table_migration_url = f"https://management.azure.com/subscriptions/{workspace_sub}/resourceGroups/{workspace_rg}/providers/Microsoft.OperationalInsights/workspaces/{workspace_name}/tables/{table_name}/migrate?api-version=2021-12-01-preview"

                send_raw_request(
                    cmd.cli_ctx,
                    method="POST",
                    url=table_migration_url)
            except Exception as e:
                raise AzureResponseError(
                    f"Old Custom Log detected. Migrating to Custom Table failed for url {table_migration_url}. Exception: {e}")
        else:
            # Define the endpoint URL
            url = f"/subscriptions/{workspace_sub}/resourceGroups/{workspace_rg}/providers/Microsoft.OperationalInsights/workspaces/{workspace_name}/tables/{table_name}?api-version=2021-12-01-preview"

            # Define the request headers
            headers = [
                'Content-Type=application/json'
            ]
            body = create_custom_table()

            try:
                send_raw_request(
                    cmd.cli_ctx,
                    method='PUT',
                    url=url,
                    headers=headers,
                    body=body)
            except Exception as e:
                raise AzureResponseError(
                    f"Creating new Custom Table failed with error {e}")

        if not dcr_found:
            # Create DCE, DCR, DCRA, AMA Agent
            dce_name = ""
            dcr_name = ""
            dcra_name = ""
            # These resources must be deployed to a Resource Group in the same
            # region as the LA workspace

            # we must do get req and loop on dce till we get an http error so we know it does not exist
            # else increase x and try again

            base_url = f"https://management.azure.com/subscriptions/{curr_subscription}/resourceGroups/{agent_rg}/providers/Microsoft.Insights/dataCollectionEndpoints/"
            api_version = "?api-version=2022-06-01"

            for index in count(start=1):
                dce_name = f"{workspace_loc}-DCE-{index}"
                dce_url = f"{base_url}{dce_name}{api_version}"
                if not does_name_exist(cmd, dce_url):
                    break

            dce_res = f"/subscriptions/{curr_subscription}/resourceGroups/{agent_rg}/providers/Microsoft.Insights/dataCollectionEndpoints/{dce_name}"

            base_url = f"https://management.azure.com/subscriptions/{curr_subscription}/resourceGroups/{agent_rg}/providers/Microsoft.Insights/dataCollectionRules/"
            api_version = "?api-version=2022-06-01"
            for index in count(start=1):
                dcr_name = f"{workspace_id}_{workspace_loc}_DCR_{index}"
                dcr_url = f"{base_url}{dcr_name}{api_version}"
                if not does_name_exist(cmd, dcr_url):
                    break

            master_template = ArmTemplateBuilder20190401()
            dce = build_dce_resource(dce_name, workspace_loc)
            master_template.add_resource(dce)
            dcr = build_dcr_resource(
                dcr_name,
                workspace_loc,
                workspace_name,
                workspace_res_id,
                dce_res,
                dce_name)
            master_template.add_resource(dcr)

            # amainstall = build_ama_install_resource(sql_virtual_machine_name, vm.location, resource_group_name, curr_subscription)
            # master_template.add_resource(amainstall)

            # /subscriptions/0009fc4d-e310-4e40-8e63-c48a23e9cdc1/resourceGroups/abhaga-iaasrg/providers/Microsoft.Insights/dataCollectionRules/0009fc4d-e310-4e40-8e63-c48a23e9cdc1_eastus_DCR_1
            dcr_resource_id = f"/subscriptions/{curr_subscription}/resourceGroups/{agent_rg}/providers/Microsoft.Insights/dataCollectionRules/{dcr_name}"

            # GET
            # https://management.azure.com/{resourceUri}/providers/Microsoft.Insights/dataCollectionRuleAssociations/{associationName}?api-version=2022-06-01

            # dcrlinkage = build_dcr_vm_linkage_resource(sql_virtual_machine_name, dcra_name, dcr_resource_id, dcr_name)

            # For DCRA do a put request after template deployment
            # PUT https://management.azure.com/subscriptions/703362b3-f278-4e4b-9179-c76eaf41ffc2/resourceGroups/myResourceGroup/providers/Microsoft.Compute/virtualMachines/myVm/providers/Microsoft.Insights/dataCollectionRuleAssociations/myAssociation?api-version=2022-06-01
            # {
            # "properties": {
            #   "dataCollectionRuleId": "/subscriptions/703362b3-f278-4e4b-9179-c76eaf41ffc2/resourceGroups/myResourceGroup/providers/Microsoft.Insights/dataCollectionRules/myCollectionRule"
            # }
            # }
            # master_template.add_resource(dcrlinkage)

            template = master_template.build()

            # deploy ARM template
            deployment_name = 'vm_deploy_' + random_string(32)
            client = get_mgmt_service_client(
                cmd.cli_ctx, ResourceType.MGMT_RESOURCE_RESOURCES).deployments
            DeploymentProperties = cmd.get_models(
                'DeploymentProperties',
                resource_type=ResourceType.MGMT_RESOURCE_RESOURCES)

            properties = DeploymentProperties(
                template=template, parameters={}, mode='incremental')

            Deployment = cmd.get_models(
                'Deployment', resource_type=ResourceType.MGMT_RESOURCE_RESOURCES)
            deployment = Deployment(properties=properties)

            # creates the AMA DEPLOYMENT
            LongRunningOperation(
                cmd.cli_ctx)(
                client.begin_create_or_update(
                    agent_rg,
                    deployment_name,
                    deployment))

            # amainstall = build_ama_install_resource(sql_virtual_machine_name, vm.location, resource_group_name, curr_subscription)
            # master_template.add_resource(amainstall)

            create_ama_and_dcra(cmd, curr_subscription, resource_group_name, sql_virtual_machine_name, workspace_id, workspace_loc, dcr_resource_id)

        else:

            # DCR and DCE were validated
            # build ARM template for linkage resource and AMA installation
            create_ama_and_dcra(cmd, curr_subscription, resource_group_name, sql_virtual_machine_name, workspace_id, workspace_loc, validated_dcr_res_id)
            return

    elif enable_assessment is False:
        # Delete DCRA
        # Otherwise AssessmentSetting payload is set above
        # GET DCRA ATTACHED TO VM: Validate for Assessment and delete
        # Unless we can track assessment dcra resource id.

        # GET
        # https://management.azure.com/subscriptions/703362b3-f278-4e4b-9179-c76eaf41ffc2/resourceGroups/myResourceGroup/providers/Microsoft.Compute/virtualMachines/myVm/providers/Microsoft.Insights/dataCollectionRuleAssociations?api-version=2022-06-01

        vm_sub = get_subscription_id(cmd.cli_ctx)
        vm_rg = resource_group_name
        vm_name = sql_virtual_machine_name

        dcra_get_url = f"https://management.azure.com/subscriptions/{vm_sub}/resourceGroups/{vm_rg}/providers/Microsoft.Compute/virtualMachines/{vm_name}/providers/Microsoft.Insights/dataCollectionRuleAssociations?api-version=2022-06-01"
        try:
            # GET on VM DCRA endpoint to list all DCRA attached to this VM
            dcra_list = send_raw_request(
                cmd.cli_ctx, method="GET", url=dcra_get_url)
        except HTTPError:
            # No DCRA Found. Assessment is disabled through AMA.
            return
        dcra_list = dcra_list.json()
        if 'value' in dcra_list and not dcra_list['value']:
            # Raise warning or message saying no DCRA found
            return
        for dcra in dcra_list['value']:

            dcra_name = dcra['name']
            pattern = re.compile(
                r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}_[a-z0-9]+_DCRA_\d+$",
                re.IGNORECASE)
            if pattern.match(dcra_name):

                # Match from response the values and add to url then delete

                curr_subscription = get_subscription_id(cmd.cli_ctx)
                resourceUri = f"subscriptions/{curr_subscription}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/virtualMachines/{sql_virtual_machine_name}"
                # DELETE
                # https://management.azure.com/{resourceUri}/providers/Microsoft.Insights/dataCollectionRuleAssociations/{associationName}?api-version=2022-06-01
                dcra_url = f"https://management.azure.com/{resourceUri}/providers/Microsoft.Insights/dataCollectionRuleAssociations/{dcra_name}?api-version=2022-06-01"
                send_raw_request(
                    cmd.cli_ctx, method="DELETE", url=dcra_url)
                return
            # Raise message DCRA deleted. Assessment disabled succesfully.

# endregion Helpers for custom commands



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
#
# Code generated by aaz-dev-tools
# --------------------------------------------------------------------------------------------

# pylint: skip-file
# flake8: noqa

from azure.cli.core.aaz import *


@register_command(
    "cosmosdb postgres cluster update",
)
class Update(AAZCommand):
    """Update an existing cluster. The request body can contain one or several properties from the cluster definition.

    Use the following reference to determine supported values for various parameters for Azure Cosmos DB for PostgreSQL clusters and nodes CLI commands: https://learn.microsoft.com/rest/api/postgresqlhsc/#values

    :example: Scale worker node compute up or down
        az cosmosdb postgres cluster update -n "test-cluster" -g "testGroup" --subscription "ffffffff-ffff-ffff-ffff-ffffffffffff" --node-v-cores 16

    :example: Scale out: Add new worker nodes
        az cosmosdb postgres cluster update -n "test-cluster" -g "testGroup" --subscription "ffffffff-ffff-ffff-ffff-ffffffffffff" --node-count 2

    :example: Scale up worker node storage
        az cosmosdb postgres cluster update -n "test-cluster" -g "testGroup" --subscription "ffffffff-ffff-ffff-ffff-ffffffffffff" --node-storage 2097152

    :example: Update multiple configuration settings of the cluster
        az cosmosdb postgres cluster update -n "test-cluster" -g "testGroup" --subscription "ffffffff-ffff-ffff-ffff-ffffffffffff" --node-v-cores 16 --node-count 4 coordinator-v-cores 16 --administrator-login-password "newPassword"

    :example: Update or define maintenance window
        az cosmosdb postgres cluster update -n "test-cluster" -g "testGroup" --subscription "ffffffff-ffff-ffff-ffff-ffffffffffff" --maintenance-window day-of-week=1 start-hour=2 --start-minute=0 custom-window="Enabled"

    :example: Upgrade PostgreSQL major version
        az cosmosdb postgres cluster update -n "test-cluster" -g "testGroup" --subscription "ffffffff-ffff-ffff-ffff-ffffffffffff" --postgresql-version "15"

    :example: Scale up coordinator storage
        az cosmosdb postgres cluster update -n "test-cluster" -g "testGroup" --subscription "ffffffff-ffff-ffff-ffff-ffffffffffff" --coordinator-storage 2097152
    """

    _aaz_info = {
        "version": "2022-11-08",
        "resources": [
            ["mgmt-plane", "/subscriptions/{}/resourcegroups/{}/providers/microsoft.dbforpostgresql/servergroupsv2/{}", "2022-11-08"],
        ]
    }

    AZ_SUPPORT_NO_WAIT = True

    def _handler(self, command_args):
        super()._handler(command_args)
        return self.build_lro_poller(self._execute_operations, self._output)

    _args_schema = None

    @classmethod
    def _build_arguments_schema(cls, *args, **kwargs):
        if cls._args_schema is not None:
            return cls._args_schema
        cls._args_schema = super()._build_arguments_schema(*args, **kwargs)

        # define Arg Group ""

        _args_schema = cls._args_schema
        _args_schema.cluster_name = AAZStrArg(
            options=["-n", "--name", "--cluster-name"],
            help="The name of the cluster.",
            required=True,
            id_part="name",
            fmt=AAZStrArgFormat(
                pattern="^(?![0-9]+$)(?!-)[a-z0-9-]{3,40}(?<!-)$",
                max_length=40,
                min_length=3,
            ),
        )
        _args_schema.resource_group = AAZResourceGroupNameArg(
            required=True,
        )

        # define Arg Group "Parameters"

        _args_schema = cls._args_schema
        _args_schema.tags = AAZDictArg(
            options=["--tags"],
            arg_group="Parameters",
            help="Application-specific metadata in the form of key-value pairs.",
        )

        tags = cls._args_schema.tags
        tags.Element = AAZStrArg()

        # define Arg Group "Properties"

        _args_schema = cls._args_schema
        _args_schema.administrator_login_password = AAZPasswordArg(
            options=["--login-password", "--administrator-login-password"],
            arg_group="Properties",
            help="The password of the administrator login. Required for creation.",
            blank=AAZPromptPasswordInput(
                msg="Password:",
            ),
        )
        _args_schema.citus_version = AAZStrArg(
            options=["--citus-version"],
            arg_group="Properties",
            help="The Citus extension version on all cluster servers.",
        )
        _args_schema.coordinator_enable_public_ip_access = AAZBoolArg(
            options=["--coord-public-ip-access", "--coordinator-enable-public-ip-access"],
            arg_group="Properties",
            help="If public access is enabled on coordinator.",
        )
        _args_schema.coordinator_server_edition = AAZStrArg(
            options=["--coord-server-edition", "--coordinator-server-edition"],
            arg_group="Properties",
            help="The edition of a coordinator server (default: GeneralPurpose). Required for creation.",
        )
        _args_schema.coordinator_storage_quota_in_mb = AAZIntArg(
            options=["--coordinator-storage", "--coordinator-storage-quota-in-mb"],
            arg_group="Properties",
            help="The storage of a server in MB. Required for creation. See https://learn.microsoft.com/rest/api/postgresqlhsc/#values for more information.",
        )
        _args_schema.coordinator_v_cores = AAZIntArg(
            options=["--coordinator-v-cores"],
            arg_group="Properties",
            help="The vCores count of a server (max: 96). Required for creation. See https://learn.microsoft.com/rest/api/postgresqlhsc/#values for more information.",
        )
        _args_schema.enable_ha = AAZBoolArg(
            options=["--enable-ha"],
            arg_group="Properties",
            help="If high availability (HA) is enabled or not for the cluster.",
        )
        _args_schema.enable_shards_on_coordinator = AAZBoolArg(
            options=["--enable-shards-on-coord", "--enable-shards-on-coordinator"],
            arg_group="Properties",
            help="If shards on coordinator is enabled or not for the cluster.",
        )
        _args_schema.maintenance_window = AAZObjectArg(
            options=["--maintenance-window"],
            arg_group="Properties",
            help="Maintenance window of a cluster.",
        )
        _args_schema.node_count = AAZIntArg(
            options=["--node-count"],
            arg_group="Properties",
            help="Worker node count of the cluster. When node count is 0, it represents a single node configuration with the ability to create distributed tables on that node. 2 or more worker nodes represent multi-node configuration. Node count value cannot be 1.",
        )
        _args_schema.node_server_edition = AAZStrArg(
            options=["--node-server-edition"],
            arg_group="Properties",
            help="The edition of a node server (default: MemoryOptimized).",
        )
        _args_schema.node_storage_quota_in_mb = AAZIntArg(
            options=["--node-storage", "--node-storage-quota-in-mb"],
            arg_group="Properties",
            help="The storage in MB on each worker node. See https://learn.microsoft.com/rest/api/postgresqlhsc/#values for more information.",
        )
        _args_schema.node_v_cores = AAZIntArg(
            options=["--node-v-cores"],
            arg_group="Properties",
            help="The compute in vCores on each worker node (max: 104). See https://learn.microsoft.com/rest/api/postgresqlhsc/#values for more information.",
        )
        _args_schema.postgresql_version = AAZStrArg(
            options=["--postgresql-version"],
            arg_group="Properties",
            help="The major PostgreSQL version on all cluster servers.",
        )
        _args_schema.preferred_primary_zone = AAZStrArg(
            options=["--preferred-primary-zone"],
            arg_group="Properties",
            help="Preferred primary availability zone (AZ) for all cluster servers.",
        )

        maintenance_window = cls._args_schema.maintenance_window
        maintenance_window.custom_window = AAZStrArg(
            options=["custom-window"],
            help="Indicates whether custom maintenance window is enabled or not.",
        )
        maintenance_window.day_of_week = AAZIntArg(
            options=["day-of-week"],
            help="Preferred day of the week for maintenance window.",
        )
        maintenance_window.start_hour = AAZIntArg(
            options=["start-hour"],
            help="Start hour within preferred day of the week for maintenance window.",
        )
        maintenance_window.start_minute = AAZIntArg(
            options=["start-minute"],
            help="Start minute within the start hour for maintenance window.",
        )
        return cls._args_schema

    def _execute_operations(self):
        self.pre_operations()
        yield self.ClustersUpdate(ctx=self.ctx)()
        self.post_operations()

    @register_callback
    def pre_operations(self):
        pass

    @register_callback
    def post_operations(self):
        pass

    def _output(self, *args, **kwargs):
        result = self.deserialize_output(self.ctx.vars.instance, client_flatten=True)
        return result

    class ClustersUpdate(AAZHttpOperation):
        CLIENT_TYPE = "MgmtClient"

        def __call__(self, *args, **kwargs):
            request = self.make_request()
            session = self.client.send_request(request=request, stream=False, **kwargs)
            if session.http_response.status_code in [202]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "location"},
                    path_format_arguments=self.url_parameters,
                )
            if session.http_response.status_code in [200]:
                return self.client.build_lro_polling(
                    self.ctx.args.no_wait,
                    session,
                    self.on_200,
                    self.on_error,
                    lro_options={"final-state-via": "location"},
                    path_format_arguments=self.url_parameters,
                )

            return self.on_error(session.http_response)

        @property
        def url(self):
            return self.client.format_url(
                "/subscriptions/{subscriptionId}/resourceGroups/{resourceGroupName}/providers/Microsoft.DBforPostgreSQL/serverGroupsv2/{clusterName}",
                **self.url_parameters
            )

        @property
        def method(self):
            return "PATCH"

        @property
        def error_format(self):
            return "MgmtErrorFormat"

        @property
        def url_parameters(self):
            parameters = {
                **self.serialize_url_param(
                    "clusterName", self.ctx.args.cluster_name,
                    required=True,
                ),
                **self.serialize_url_param(
                    "resourceGroupName", self.ctx.args.resource_group,
                    required=True,
                ),
                **self.serialize_url_param(
                    "subscriptionId", self.ctx.subscription_id,
                    required=True,
                ),
            }
            return parameters

        @property
        def query_parameters(self):
            parameters = {
                **self.serialize_query_param(
                    "api-version", "2022-11-08",
                    required=True,
                ),
            }
            return parameters

        @property
        def header_parameters(self):
            parameters = {
                **self.serialize_header_param(
                    "Content-Type", "application/json",
                ),
                **self.serialize_header_param(
                    "Accept", "application/json",
                ),
            }
            return parameters

        @property
        def content(self):
            _content_value, _builder = self.new_content_builder(
                self.ctx.args,
                typ=AAZObjectType,
                typ_kwargs={"flags": {"required": True, "client_flatten": True}}
            )
            _builder.set_prop("properties", AAZObjectType, typ_kwargs={"flags": {"client_flatten": True}})
            _builder.set_prop("tags", AAZDictType, ".tags")

            properties = _builder.get(".properties")
            if properties is not None:
                properties.set_prop("administratorLoginPassword", AAZStrType, ".administrator_login_password", typ_kwargs={"flags": {"secret": True}})
                properties.set_prop("citusVersion", AAZStrType, ".citus_version")
                properties.set_prop("coordinatorEnablePublicIpAccess", AAZBoolType, ".coordinator_enable_public_ip_access")
                properties.set_prop("coordinatorServerEdition", AAZStrType, ".coordinator_server_edition")
                properties.set_prop("coordinatorStorageQuotaInMb", AAZIntType, ".coordinator_storage_quota_in_mb")
                properties.set_prop("coordinatorVCores", AAZIntType, ".coordinator_v_cores")
                properties.set_prop("enableHa", AAZBoolType, ".enable_ha")
                properties.set_prop("enableShardsOnCoordinator", AAZBoolType, ".enable_shards_on_coordinator")
                properties.set_prop("maintenanceWindow", AAZObjectType, ".maintenance_window")
                properties.set_prop("nodeCount", AAZIntType, ".node_count")
                properties.set_prop("nodeServerEdition", AAZStrType, ".node_server_edition")
                properties.set_prop("nodeStorageQuotaInMb", AAZIntType, ".node_storage_quota_in_mb")
                properties.set_prop("nodeVCores", AAZIntType, ".node_v_cores")
                properties.set_prop("postgresqlVersion", AAZStrType, ".postgresql_version")
                properties.set_prop("preferredPrimaryZone", AAZStrType, ".preferred_primary_zone")

            maintenance_window = _builder.get(".properties.maintenanceWindow")
            if maintenance_window is not None:
                maintenance_window.set_prop("customWindow", AAZStrType, ".custom_window")
                maintenance_window.set_prop("dayOfWeek", AAZIntType, ".day_of_week")
                maintenance_window.set_prop("startHour", AAZIntType, ".start_hour")
                maintenance_window.set_prop("startMinute", AAZIntType, ".start_minute")

            tags = _builder.get(".tags")
            if tags is not None:
                tags.set_elements(AAZStrType, ".")

            return self.serialize_content(_content_value)

        def on_200(self, session):
            data = self.deserialize_http_content(session)
            self.ctx.set_var(
                "instance",
                data,
                schema_builder=self._build_schema_on_200
            )

        _schema_on_200 = None

        @classmethod
        def _build_schema_on_200(cls):
            if cls._schema_on_200 is not None:
                return cls._schema_on_200

            cls._schema_on_200 = AAZObjectType()
            _UpdateHelper._build_schema_cluster_read(cls._schema_on_200)

            return cls._schema_on_200


class _UpdateHelper:
    """Helper class for Update"""

    _schema_cluster_read = None

    @classmethod
    def _build_schema_cluster_read(cls, _schema):
        if cls._schema_cluster_read is not None:
            _schema.id = cls._schema_cluster_read.id
            _schema.location = cls._schema_cluster_read.location
            _schema.name = cls._schema_cluster_read.name
            _schema.properties = cls._schema_cluster_read.properties
            _schema.system_data = cls._schema_cluster_read.system_data
            _schema.tags = cls._schema_cluster_read.tags
            _schema.type = cls._schema_cluster_read.type
            return

        cls._schema_cluster_read = _schema_cluster_read = AAZObjectType()

        cluster_read = _schema_cluster_read
        cluster_read.id = AAZStrType(
            flags={"read_only": True},
        )
        cluster_read.location = AAZStrType(
            flags={"required": True},
        )
        cluster_read.name = AAZStrType(
            flags={"read_only": True},
        )
        cluster_read.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        cluster_read.system_data = AAZObjectType(
            serialized_name="systemData",
            flags={"read_only": True},
        )
        cls._build_schema_system_data_read(cluster_read.system_data)
        cluster_read.tags = AAZDictType()
        cluster_read.type = AAZStrType(
            flags={"read_only": True},
        )

        properties = _schema_cluster_read.properties
        properties.administrator_login = AAZStrType(
            serialized_name="administratorLogin",
            flags={"read_only": True},
        )
        properties.citus_version = AAZStrType(
            serialized_name="citusVersion",
        )
        properties.coordinator_enable_public_ip_access = AAZBoolType(
            serialized_name="coordinatorEnablePublicIpAccess",
        )
        properties.coordinator_server_edition = AAZStrType(
            serialized_name="coordinatorServerEdition",
        )
        properties.coordinator_storage_quota_in_mb = AAZIntType(
            serialized_name="coordinatorStorageQuotaInMb",
        )
        properties.coordinator_v_cores = AAZIntType(
            serialized_name="coordinatorVCores",
        )
        properties.earliest_restore_time = AAZStrType(
            serialized_name="earliestRestoreTime",
            flags={"read_only": True},
        )
        properties.enable_ha = AAZBoolType(
            serialized_name="enableHa",
        )
        properties.enable_shards_on_coordinator = AAZBoolType(
            serialized_name="enableShardsOnCoordinator",
        )
        properties.maintenance_window = AAZObjectType(
            serialized_name="maintenanceWindow",
        )
        properties.node_count = AAZIntType(
            serialized_name="nodeCount",
        )
        properties.node_enable_public_ip_access = AAZBoolType(
            serialized_name="nodeEnablePublicIpAccess",
        )
        properties.node_server_edition = AAZStrType(
            serialized_name="nodeServerEdition",
        )
        properties.node_storage_quota_in_mb = AAZIntType(
            serialized_name="nodeStorageQuotaInMb",
        )
        properties.node_v_cores = AAZIntType(
            serialized_name="nodeVCores",
        )
        properties.point_in_time_utc = AAZStrType(
            serialized_name="pointInTimeUTC",
        )
        properties.postgresql_version = AAZStrType(
            serialized_name="postgresqlVersion",
        )
        properties.preferred_primary_zone = AAZStrType(
            serialized_name="preferredPrimaryZone",
        )
        properties.private_endpoint_connections = AAZListType(
            serialized_name="privateEndpointConnections",
            flags={"read_only": True},
        )
        properties.provisioning_state = AAZStrType(
            serialized_name="provisioningState",
            flags={"read_only": True},
        )
        properties.read_replicas = AAZListType(
            serialized_name="readReplicas",
            flags={"read_only": True},
        )
        properties.server_names = AAZListType(
            serialized_name="serverNames",
            flags={"read_only": True},
        )
        properties.source_location = AAZStrType(
            serialized_name="sourceLocation",
        )
        properties.source_resource_id = AAZStrType(
            serialized_name="sourceResourceId",
        )
        properties.state = AAZStrType(
            flags={"read_only": True},
        )

        maintenance_window = _schema_cluster_read.properties.maintenance_window
        maintenance_window.custom_window = AAZStrType(
            serialized_name="customWindow",
        )
        maintenance_window.day_of_week = AAZIntType(
            serialized_name="dayOfWeek",
        )
        maintenance_window.start_hour = AAZIntType(
            serialized_name="startHour",
        )
        maintenance_window.start_minute = AAZIntType(
            serialized_name="startMinute",
        )

        private_endpoint_connections = _schema_cluster_read.properties.private_endpoint_connections
        private_endpoint_connections.Element = AAZObjectType()

        _element = _schema_cluster_read.properties.private_endpoint_connections.Element
        _element.id = AAZStrType(
            flags={"read_only": True},
        )
        _element.name = AAZStrType(
            flags={"read_only": True},
        )
        _element.properties = AAZObjectType(
            flags={"client_flatten": True},
        )
        _element.system_data = AAZObjectType(
            serialized_name="systemData",
            flags={"read_only": True},
        )
        cls._build_schema_system_data_read(_element.system_data)
        _element.type = AAZStrType(
            flags={"read_only": True},
        )

        properties = _schema_cluster_read.properties.private_endpoint_connections.Element.properties
        properties.group_ids = AAZListType(
            serialized_name="groupIds",
        )
        properties.private_endpoint = AAZObjectType(
            serialized_name="privateEndpoint",
        )
        properties.private_link_service_connection_state = AAZObjectType(
            serialized_name="privateLinkServiceConnectionState",
        )

        group_ids = _schema_cluster_read.properties.private_endpoint_connections.Element.properties.group_ids
        group_ids.Element = AAZStrType()

        private_endpoint = _schema_cluster_read.properties.private_endpoint_connections.Element.properties.private_endpoint
        private_endpoint.id = AAZStrType()

        private_link_service_connection_state = _schema_cluster_read.properties.private_endpoint_connections.Element.properties.private_link_service_connection_state
        private_link_service_connection_state.actions_required = AAZStrType(
            serialized_name="actionsRequired",
        )
        private_link_service_connection_state.description = AAZStrType()
        private_link_service_connection_state.status = AAZStrType()

        read_replicas = _schema_cluster_read.properties.read_replicas
        read_replicas.Element = AAZStrType()

        server_names = _schema_cluster_read.properties.server_names
        server_names.Element = AAZObjectType()

        _element = _schema_cluster_read.properties.server_names.Element
        _element.fully_qualified_domain_name = AAZStrType(
            serialized_name="fullyQualifiedDomainName",
            flags={"read_only": True},
        )
        _element.name = AAZStrType()

        tags = _schema_cluster_read.tags
        tags.Element = AAZStrType()

        _schema.id = cls._schema_cluster_read.id
        _schema.location = cls._schema_cluster_read.location
        _schema.name = cls._schema_cluster_read.name
        _schema.properties = cls._schema_cluster_read.properties
        _schema.system_data = cls._schema_cluster_read.system_data
        _schema.tags = cls._schema_cluster_read.tags
        _schema.type = cls._schema_cluster_read.type

    _schema_system_data_read = None

    @classmethod
    def _build_schema_system_data_read(cls, _schema):
        if cls._schema_system_data_read is not None:
            _schema.created_at = cls._schema_system_data_read.created_at
            _schema.created_by = cls._schema_system_data_read.created_by
            _schema.created_by_type = cls._schema_system_data_read.created_by_type
            _schema.last_modified_at = cls._schema_system_data_read.last_modified_at
            _schema.last_modified_by = cls._schema_system_data_read.last_modified_by
            _schema.last_modified_by_type = cls._schema_system_data_read.last_modified_by_type
            return

        cls._schema_system_data_read = _schema_system_data_read = AAZObjectType(
            flags={"read_only": True}
        )

        system_data_read = _schema_system_data_read
        system_data_read.created_at = AAZStrType(
            serialized_name="createdAt",
        )
        system_data_read.created_by = AAZStrType(
            serialized_name="createdBy",
        )
        system_data_read.created_by_type = AAZStrType(
            serialized_name="createdByType",
        )
        system_data_read.last_modified_at = AAZStrType(
            serialized_name="lastModifiedAt",
        )
        system_data_read.last_modified_by = AAZStrType(
            serialized_name="lastModifiedBy",
        )
        system_data_read.last_modified_by_type = AAZStrType(
            serialized_name="lastModifiedByType",
        )

        _schema.created_at = cls._schema_system_data_read.created_at
        _schema.created_by = cls._schema_system_data_read.created_by
        _schema.created_by_type = cls._schema_system_data_read.created_by_type
        _schema.last_modified_at = cls._schema_system_data_read.last_modified_at
        _schema.last_modified_by = cls._schema_system_data_read.last_modified_by
        _schema.last_modified_by_type = cls._schema_system_data_read.last_modified_by_type


__all__ = ["Update"]
